import ssl
import json
import atexit
import logging
from typing import Dict, Optional

from grpclib.client import Channel
from grpclib.config import Configuration
from grpclib.events import SendMessage, SendRequest, RecvInitialMetadata, listen

from qm.api.models.debug_data import DebugData

logger = logging.getLogger(__name__)


def _create_debug_data_event(debug_data: DebugData, channel: Channel) -> None:
    async def intercept_response(event: RecvInitialMetadata) -> None:
        metadata = event.metadata
        logger.debug(f"Collected response metadata: {json.dumps(dict(metadata), indent=4)}")
        debug_data.append(metadata)

    async def send_request_debug(event: SendRequest) -> None:
        logger.debug("-----------request start-----------")
        logger.debug("   ---    request headers    ---   ")
        logger.debug(f"method:       {event.method_name}")
        logger.debug(f"metadata:     {json.dumps(dict(event.metadata), indent=4)}")
        logger.debug(f"content type: {event.content_type}")
        if event.deadline:
            deadline = event.deadline.time_remaining()
        else:
            deadline = None
        logger.debug(f"deadline:     {deadline}")

    async def send_message_debug(event: SendMessage) -> None:
        logger.debug("   ---    request message    ---   ")
        try:
            logger.debug(f"message:      {event.message.to_json(4)}")
        except TypeError:
            pass
        logger.debug("------------end request------------")

    listen(channel, RecvInitialMetadata, intercept_response)
    listen(channel, SendRequest, send_request_debug)
    listen(channel, SendMessage, send_message_debug)


def _create_add_headers_event(headers: Dict[str, str], channel: Channel) -> None:
    async def add_headers(event: SendRequest) -> None:
        event.metadata.update(headers)

    listen(channel, SendRequest, add_headers)


async def create_channel(
    host: str,
    port: int,
    ssl_context: Optional[ssl.SSLContext],
    max_message_size: int,
    headers: Dict[str, str],
    debug_data: Optional[DebugData] = None,
) -> Channel:

    channel = Channel(
        host=host,
        port=port,
        ssl=ssl_context,
        config=Configuration(
            http2_connection_window_size=max_message_size,
            http2_stream_window_size=max_message_size,
        ),
    )

    if debug_data:
        _create_debug_data_event(debug_data, channel)

    _create_add_headers_event(headers, channel)

    atexit.register(channel.close)

    return channel



import os
import pathlib
import dataclasses
from typing import Any, Union, TypeVar, Protocol, runtime_checkable

import numpy
import numpy.typing as npt

PathLike = Union[str, bytes, pathlib.Path, os.PathLike]  # type: ignore[type-arg]
Number = Union[int, float]
Value = Union[Number, bool]
NumberT = TypeVar("NumberT", bool, int, float)
"""A generic type variable that can be used to represent a pythonic `int`, `bool` or `float`."""

NumpyNumber = Union[numpy.floating[Any], numpy.integer[Any]]
NumpyValue = Union[NumpyNumber, numpy.bool_]
NumpyArray = npt.NDArray[numpy.generic]

NumpySupportedNumber = Union[Number, NumpyNumber]
NumpySupportedFloat = Union[float, numpy.floating[Any]]
NumpySupportedValue = Union[Value, NumpyValue]


@runtime_checkable
@dataclasses.dataclass
class DataclassProtocol(Protocol):
    pass



import warnings
from collections.abc import Iterable
from typing import Union, Optional, Sequence

from qm.type_hinting import Number
from qm.qua._dsl._type_hints import OneOrMore
from qm.qua._dsl.stream_processing.map_functions.map_function_classes import (
    FFT,
    Demod,
    Average,
    DotProduct,
    Convolution,
    BooleanToInt,
    TupleMultiply,
    ConvolutionMode,
    TupleDotProduct,
    MultiplyByScalar,
    MultiplyByVector,
    TupleConvolution,
)


class MapFunctions:
    @staticmethod
    def average(axis: Optional[OneOrMore[Number]] = None) -> Average:
        """Perform a running average on a stream item. The Output of this operation is the
        running average of the values in the stream starting from the beginning of the
        QUA program.

        Args:
            axis: optional Axis or axes along which to average.

        Returns:
            stream object
        """
        return Average(axis)

    @staticmethod
    def dot_product(vector: Sequence[Number]) -> DotProduct:
        """Computes dot product of the given vector and an item of the input stream

        Args:
            vector: constant vector of numbers

        Returns:
            stream object
        """
        return DotProduct(vector)

    @staticmethod
    def tuple_dot_product() -> TupleDotProduct:
        """Computes dot product between the two vectors of the input stream

        Returns:
            stream object
        """
        return TupleDotProduct()

    @staticmethod
    def multiply_by(scalar_or_vector: OneOrMore[Number]) -> Union[MultiplyByVector, MultiplyByScalar]:
        """Multiply the input stream item by a constant scalar or vector.
        the input item can be either scalar or vector.

        Args:
            scalar_or_vector: either a scalar number, or a vector of
                scalars.

        Returns:
            stream object
        """
        if isinstance(scalar_or_vector, Iterable):
            return MultiplyByVector(scalar_or_vector)
        else:
            return MultiplyByScalar(scalar_or_vector)

    @staticmethod
    def tuple_multiply() -> TupleMultiply:
        """Computes multiplication between the two elements of the input stream.
        Can be any combination of scalar and vectors.

        Returns:
            stream object
        """
        return TupleMultiply()

    @staticmethod
    def convolution(constant_vector: Sequence[Number], mode: Optional[ConvolutionMode] = "") -> Convolution:
        """Computes discrete, linear convolution of one-dimensional constant vector and
        one-dimensional vector item of the input stream.

        Args:
            constant_vector: vector of numbers
            mode: "full", "same" or "valid"

        Returns:
            stream object
        """
        if mode is None:
            warnings.warn(
                "mode=None is deprecated, use empty-string or (recommended) don't write the mode at-all.",
                DeprecationWarning,
            )
            mode = ""
        return Convolution(constant_vector, mode)

    @staticmethod
    def tuple_convolution(mode: Optional[ConvolutionMode] = "") -> TupleConvolution:
        """Computes discrete, linear convolution of two one-dimensional vectors of the
        input stream

        Args:
            mode: "full", "same" or "valid"

        Returns:
            stream object
        """
        if mode is None:
            warnings.warn(
                "mode=None is deprecated, use empty-string or (recommended) don't write the mode at-all.",
                DeprecationWarning,
            )
            mode = ""
        return TupleConvolution(mode)

    @staticmethod
    def fft(output: Optional[str] = None) -> FFT:
        """Computes one-dimensional discrete fourier transform for every item in the
        stream.
        Item can be a vector of numbers, in this case fft will assume all imaginary
        numbers are 0.
        Item can also be a vector of number pairs - in this case for each pair - the
        first will be real and second imaginary.

        Args:
            output: supported from QOP 1.30 and QOP 2.0, options are
                "normal", "abs" and "angle":

                *   "normal" - Same as default (none), returns a 2d array of
                    size Nx2, where N is the length of the original vector.
                    The first item in each pair is the real part, and the 2nd
                    is the imaginary part.
                *   "abs" - Returns a 1d array of size N with the abs of the fft.
                *   "angle" - Returns the angle between the imaginary and real
                    parts in radians.

        Returns:
            stream object
        """
        return FFT(output)

    @staticmethod
    def boolean_to_int() -> BooleanToInt:
        """
        Converts boolean to integer number - 1 for true and 0 for false

        Returns:
            stream object
        """
        return BooleanToInt()

    @staticmethod
    def demod(
        frequency: Number,
        iw_cos: OneOrMore[Number],
        iw_sin: OneOrMore[Number],
        *,
        integrate: Optional[bool] = None,
    ) -> Demod:
        """Demodulates the acquired data from the indicated stream at the given frequency
        and integration weights.
        If operating on a stream of tuples, assumes that the 2nd item is the timestamps
        and uses them for the demodulation, reproducing the demodulation performed
        in real time.
        If operated on a single stream, assumes that the first item is at time zero and
        that the elements are separated by 1ns.

        Args:
            frequency: frequency for demodulation calculation
            iw_cos: cosine integration weight. Integration weight can be
                either a scalar for constant integration weight, or a
                python iterable for arbitrary integration weights.
            iw_sin: sine integration weight. Integration weight can be
                either a scalar for constant integration weight, or a
                python iterable for arbitrary integration weights.
            integrate: sum the demodulation result and returns a scalar
                if True (default), else the demodulated stream without
                summation is returned

        Returns:
            stream object

        Example:
            ```python
            with stream_processing():
                adc_stream.input1().with_timestamps().map(FUNCTIONS.demod(freq, 1.0, 0.0, integrate=False)).average().save('cos_env')
                adc_stream.input1().with_timestamps().map(FUNCTIONS.demod(freq, 1.0, 0.0)).average().save('cos_result')  # Default is integrate=True
            ```

        Note:
            The demodulation in the stream processing **does not** take in consideration
            any real-time modifications to the frame, phase or frequency of the element.
            If the program has any QUA command that changes them, the result of the
            stream processing demodulation will be invalid.

        """
        return Demod(frequency, iw_cos, iw_sin, integrate)


FUNCTIONS = MapFunctions()



from typing import Tuple, Optional, overload

from qm.exceptions import QmQuaException
from qm.grpc.qua import QuaProgramAnyScalarExpression
from qm.qua._dsl._type_hints import MessageExpressionType
from qm.qua._expressions import Scalar, to_scalar_pb_expression

AmpValuesType = Tuple[
    MessageExpressionType,
    Optional[MessageExpressionType],
    Optional[MessageExpressionType],
    Optional[MessageExpressionType],
]


# Although _PulseAmp is a protected class, it is used by QUAM.
# Therefore, any breaking changes should be communicated to the QUAM authors.
class _PulseAmp:
    def __init__(
        self,
        v1: QuaProgramAnyScalarExpression,
        v2: Optional[QuaProgramAnyScalarExpression],
        v3: Optional[QuaProgramAnyScalarExpression],
        v4: Optional[QuaProgramAnyScalarExpression],
    ):
        if v1 is None:
            raise QmQuaException("amp can be one value or a matrix of 4")
        if v2 is None and v3 is None and v4 is None:
            pass
        elif v2 is not None and v3 is not None and v4 is not None:
            pass
        else:
            raise QmQuaException("amp can be one value or a matrix of 4.")

        self.v1 = v1
        self.v2 = v2
        self.v3 = v3
        self.v4 = v4

    def value(self) -> AmpValuesType:
        return self.v1, self.v2, self.v3, self.v4

    def __rmul__(self, other: str) -> Tuple[str, AmpValuesType]:
        return self * other

    def __mul__(self, other: str) -> Tuple[str, AmpValuesType]:
        if not isinstance(other, str):
            raise QmQuaException("you can multiply only a pulse")
        return other, self.value()


@overload
def amp(v1: Scalar[float]) -> _PulseAmp:
    ...


@overload
def amp(v1: Scalar[float], v2: Scalar[float], v3: Scalar[float], v4: Scalar[float]) -> _PulseAmp:
    ...


def amp(
    v1: Scalar[float],
    v2: Optional[Scalar[float]] = None,
    v3: Optional[Scalar[float]] = None,
    v4: Optional[Scalar[float]] = None,
) -> _PulseAmp:
    """To be used only within a [play][qm.qua.play] or [measure][qm.qua.measure] command, as a multiplication to
    the `operation`.

    It is possible to scale the pulse's amplitude dynamically by using the following syntax:

    ``play('pulse_name' * amp(v), 'element')``

    where ``v`` is QUA variable of type fixed. Range of v: -2 to $2 - 2^{-16}$ in steps of $2^{-16}$.

    Moreover, if the pulse is intended to a mixedInputs element and thus is defined with two waveforms,
    the two waveforms, described as a column vector, can be multiplied by a matrix:

    ``play('pulse_name' * amp(v_00, v_01, v_10, v_11), 'element'),``

    where ``v_ij``, i,j={0,1}, are QUA variables of type fixed.
    Note that ``v_ij`` should satisfy -2 <= ``v_ij`` <= $2 - 2{-16}$.

    Note that scaling in this manner, rather than in the configuration, might result
    in a computational overhead.
    See [QUA Best Practice Guide](../../Guides/best_practices.md#general) for more information.

    Args:
        v1: If only this variable is given, it is the scaler amplitude
            factor which multiples the `pulse` associated with the
            `operation`. If all variables are given, then it is the
            first element in the amplitude matrix which multiples the
            `pulse` associated with the `operation`.
        v2: The second element in the amplitude matrix which multiples
            the `pulse` associated with the `operation`.
        v3: The third element in the amplitude matrix which multiples
            the `pulse` associated with the `operation`.
        v4: The forth element in the amplitude matrix which multiples
            the `pulse` associated with the `operation`.
    """

    def _cast_number(v: Optional[Scalar[float]]) -> Optional[QuaProgramAnyScalarExpression]:
        if v is None:
            return None
        return to_scalar_pb_expression(v)

    return _PulseAmp(to_scalar_pb_expression(v1), _cast_number(v2), _cast_number(v3), _cast_number(v4))



from enum import Enum
from typing import Union, Literal, overload

from qm._loc import _get_loc
from qm.exceptions import QmQuaException
from qm.api.models.capabilities import QopCaps
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._expressions import StructT, QuaExternalIncomingStream, QuaExternalOutgoingStream
from qm.grpc.qua import (
    QuaProgramAnyStatement,
    QuaProgramSendToExternalStreamStatement,
    QuaProgramReceiveFromExternalStreamStatement,
)


class QuaStreamDirection(Enum):
    INCOMING = "INCOMING"
    OUTGOING = "OUTGOING"


@overload
def declare_external_stream(
    struct_type: type[StructT], stream_id: int, direction: Literal[QuaStreamDirection.INCOMING]
) -> QuaExternalIncomingStream[StructT]:
    ...


@overload
def declare_external_stream(
    struct_type: type[StructT], stream_id: int, direction: Literal[QuaStreamDirection.OUTGOING]
) -> QuaExternalOutgoingStream[StructT]:
    ...


def declare_external_stream(
    struct_type: type[StructT], stream_id: int, direction: QuaStreamDirection
) -> Union[QuaExternalIncomingStream[StructT], QuaExternalOutgoingStream[StructT]]:
    """Declare a stream to an external compute resource.

    The stream can be either an incoming or an outgoing stream, which will receive or send data from the external
    compute resource respectively.

    Note:
        Alpha version.
        This function may change its signature and behavior in the future without obeying semantic
        versioning and without maintaining backward compatibility.

    Args:
        struct_type: A QuaStruct type that defines a single packet of the stream.
        stream_id: The ID of the stream, an integer between 0 and 1000.
            Has to match the ID of the stream declared in
            the external compute resource.
        direction: The direction of the stream, either QuaStreamDirection.INCOMING or QuaStreamDirection.OUTGOING.

    Example:
        ```python
        struct = declare_struct(StructType)
        stream = declare_external_stream(StructType, 0, QuaStreamDirection.INCOMING)
        ```

    """
    program_scope = scopes_manager.program_scope
    program_scope.add_used_capability(QopCaps.external_stream)

    if not isinstance(stream_id, int):
        raise QmQuaException("stream_id must be an integer")

    if not 1 <= stream_id <= 1023:
        raise QmQuaException("stream_id must be between 1 and 1023")

    stream: Union[QuaExternalIncomingStream[StructT], QuaExternalOutgoingStream[StructT]]

    stream_identifier = (stream_id, direction)
    if stream_identifier in program_scope.declared_external_streams:
        raise QmQuaException("external stream already declared")

    if direction == QuaStreamDirection.INCOMING:
        stream = QuaExternalIncomingStream(stream_id, struct_type)
    elif direction == QuaStreamDirection.OUTGOING:
        stream = QuaExternalOutgoingStream(stream_id, struct_type)
    else:
        raise QmQuaException("direction must be either QuaStreamDirection.INCOMING or QuaStreamDirection.OUTGOING")

    program_scope.add_external_stream_declaration(stream_identifier, stream.declaration_statement)
    return stream


def send_to_external_stream(stream: QuaExternalOutgoingStream[StructT], struct: StructT) -> None:
    """
    Send data to an external compute resource.

    Note:
        Alpha version.
        This function may change its signature and behavior in the future without obeying semantic
        versioning and without maintaining backward compatibility.

    Args:
        stream: The outgoing stream to send data to.
        struct: The struct containing the data to send.
    """
    statement = QuaProgramAnyStatement(
        send_to_external_stream=QuaProgramSendToExternalStreamStatement(
            loc=_get_loc(), stream=stream.unwrapped, struct=struct.struct_reference.unwrapped
        )
    )
    scopes_manager.append_statement(statement)


def receive_from_external_stream(stream: QuaExternalIncomingStream[StructT], struct: StructT) -> None:
    """
    Receive data from an external compute resource.

    Note:
        Alpha version.
        This function may change its signature and behavior in the future without obeying semantic
        versioning and without maintaining backward compatibility.

    Args:
        stream: The incoming stream to receive data from.
        struct: The struct to store the received data in.
    """
    statement = QuaProgramAnyStatement(
        receive_from_external_stream=QuaProgramReceiveFromExternalStreamStatement(
            loc=_get_loc(), stream=stream.unwrapped, struct=struct.struct_reference.unwrapped
        )
    )
    scopes_manager.append_statement(statement)



import logging
from io import BytesIO
from collections import defaultdict
from typing import Dict, Tuple, Union, Literal, Mapping, BinaryIO, Optional, Collection

from qm.utils.async_utils import run_async
from qm.type_hinting.general import NumpyArray
from qm.api.v2.job_result_api import JobResultApi
from qm.api.models.jobs import JobResultItemSchema, JobNamedResultHeader
from qm._stream_results._utils import (
    _standardize_slice,
    assert_no_dataloss,
    log_execution_errors,
    _create_results_array,
)

logger = logging.getLogger(__name__)


class MultipleStreamsFetcher:
    def __init__(
        self,
        schemas: Mapping[str, JobResultItemSchema],
        service: JobResultApi,
    ) -> None:
        self._schemas = schemas
        self._service = service

    @property
    def _job_id(self) -> str:
        return self._service.id

    def strict_fetch(
        self,
        items: Mapping[str, Union[int, slice]],
        flat_struct_items: Union[Collection[str], Literal["all"]] = frozenset(),
        check_for_errors: bool = True,
        timeout: Optional[float] = None,
    ) -> Mapping[str, NumpyArray]:

        headers = self._get_named_headers(items, flat_struct_items)
        name_to_slice, name_to_header = self._standardize_query_params(items, headers, check_for_errors)
        return self._fetch_results(name_to_slice, name_to_header, timeout=timeout)

    def _get_named_headers(
        self, items: Collection[str], flat_struct_items: Union[Collection[str], Literal["all"]]
    ) -> Mapping[str, JobNamedResultHeader]:
        if flat_struct_items == "all":
            flat_struct_items = frozenset(items)
        return self._service.get_named_headers({name: (name in flat_struct_items) for name in items})

    def _fetch_results(
        self,
        name_to_slice: Mapping[str, slice],
        name_to_header: Mapping[str, JobNamedResultHeader],
        timeout: Optional[float],
    ) -> Mapping[str, NumpyArray]:
        name_to_writer = {n: BytesIO() for n in name_to_slice}
        name_to_count_data_written = run_async(
            self._add_results_to_writers(name_to_slice, name_to_writer, timeout=timeout)
        )
        name_to_array = {
            n: _create_results_array(
                name_to_count_data_written[n],
                name_to_header[n],
                name_to_writer[n],
            )
            for n in name_to_slice
        }
        return name_to_array

    def _get_and_validate_header(
        self, name: str, headers: Mapping[str, JobNamedResultHeader], check_for_errors: bool = True
    ) -> JobNamedResultHeader:
        if name not in headers:
            raise Exception(f"Result named '{name}' not found for job {self._job_id}")
        header = headers[name]
        log_execution_errors(header, name, check_for_errors)
        assert_no_dataloss(header, self._job_id)
        return header

    def _standardize_query_params(
        self,
        items: Mapping[str, Union[int, slice]],
        headers: Mapping[str, JobNamedResultHeader],
        check_for_errors: bool = True,
    ) -> Tuple[Mapping[str, slice], Mapping[str, JobNamedResultHeader]]:
        name_to_slice = {}
        name_to_header = {}
        for name, item in items.items():
            header = self._get_and_validate_header(name, headers, check_for_errors)
            slicer = _standardize_slice(name, item, header)
            name_to_slice[name] = slicer
            name_to_header[name] = header
        return name_to_slice, name_to_header

    async def _add_results_to_writers(
        self,
        name_to_slice: Mapping[str, slice],
        name_to_writer: Mapping[str, BinaryIO],
        timeout: Optional[float],
    ) -> Mapping[str, int]:
        name_to_count_data_written: Dict[str, int] = defaultdict(int)
        async for result in self._service.get_job_named_results(name_to_slice, timeout=timeout):
            data_writer = name_to_writer[result.output_name]
            data_writer.write(result.data)
            name_to_count_data_written[result.output_name] += result.count_of_items

        return name_to_count_data_written



from dependency_injector.wiring import Provide, inject

from qm.grpc.qua_config import QuaConfig
from qm.type_hinting.config_types import FullQuaConfig
from qm.api.models.capabilities import ServerCapabilities
from qm.containers.capabilities_container import CapabilitiesContainer

from ._dict_to_pb_converter import DictToQuaConfigConverter


@inject
def convert_msg_to_config(
    config: QuaConfig,
    capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities],
) -> FullQuaConfig:
    converter = DictToQuaConfigConverter(capabilities)
    return converter.deconvert(config)



import logging
from typing import List, Type, Mapping, Optional, AsyncIterator

from qm.api.base_api import BaseApi
from qm.api.models.server_details import ConnectionDetails
from qm.api.models.jobs import JobNamedResult, JobStreamingState, JobResultItemSchema, JobNamedResultHeader
from qm.grpc.results_analyser import (
    GetJobStateRequest,
    GetJobErrorsRequest,
    JobResultsServiceStub,
    GetJobNamedResultRequest,
    GetJobErrorsResponseError,
    GetJobResultSchemaRequest,
    GetJobNamedResultHeaderRequest,
)

logger = logging.getLogger(__name__)


class JobResultServiceApi(BaseApi[JobResultsServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str):
        super().__init__(connection_details)
        self._id = job_id

    @property
    def id(self) -> str:
        return self._id

    @property
    def _stub_class(self) -> Type[JobResultsServiceStub]:
        return JobResultsServiceStub

    def get_job_errors(self) -> List[GetJobErrorsResponseError]:
        request = GetJobErrorsRequest(job_id=self._id)
        response = self._run(self._stub.get_job_errors(request, timeout=self._timeout))
        return response.errors

    async def get_job_named_result(
        self,
        output_name: str,
        long_offset: int,
        limit: int,
        timeout: Optional[float],
    ) -> AsyncIterator[JobNamedResult]:
        timeout = timeout if timeout is not None else self._timeout

        request = GetJobNamedResultRequest(
            job_id=self._id, output_name=output_name, long_offset=long_offset, limit=limit
        )
        response = self._run_async_iterator(self._stub.get_job_named_result, request, timeout=timeout)

        async for result in response:
            yield JobNamedResult(data=result.data, count_of_items=result.count_of_items, output_name=output_name)

    def get_job_state(self) -> JobStreamingState:
        request = GetJobStateRequest(job_id=self._id)
        response = self._run(self._stub.get_job_state(request, timeout=self._timeout))
        return JobStreamingState(
            job_id=self._id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.has_dataloss,
        )

    def get_named_header(self, output_name: str, flat_struct: bool) -> JobNamedResultHeader:
        request = GetJobNamedResultHeaderRequest(job_id=self._id, output_name=output_name, flat_format=flat_struct)
        response = self._run(self._stub.get_job_named_result_header(request, timeout=self._timeout))
        return JobNamedResultHeader(
            count_so_far=response.count_so_far,
            bare_dtype=response.simple_d_type,
            shape=tuple(response.shape),
            has_dataloss=response.has_dataloss,
            has_execution_errors=bool(response.has_execution_errors),
        )

    def get_state_from_header(self, output_name: str, flat_struct: bool) -> JobStreamingState:
        """This function is for backward compatibility, when we didn't have the get_job_state"""
        request = GetJobNamedResultHeaderRequest(job_id=self._id, output_name=output_name, flat_format=flat_struct)
        response = self._run(self._stub.get_job_named_result_header(request, timeout=self._timeout))
        return JobStreamingState(
            job_id=self._id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.has_dataloss,
        )

    def get_job_result_schema(self) -> Mapping[str, JobResultItemSchema]:
        request = GetJobResultSchemaRequest(job_id=self._id)
        response = self._run(self._stub.get_job_result_schema(request, timeout=self._timeout))
        return {
            item.name: JobResultItemSchema(
                name=item.name,
                bare_dtype=item.simple_d_type,
                shape=tuple(item.shape),
                is_single=item.is_single,
                expected_count=item.expected_count,
            )
            for item in response.items
        }



import logging
import os.path
import dataclasses
from copy import deepcopy
from dataclasses import dataclass
from collections import defaultdict
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List, Type, Union, Mapping, TypeVar, Callable, Optional, Protocol, Sequence, cast

from qm.simulate import SimulatorSamples
from qm.waveform_report._utils import format_float, pretty_string_freq
from qm.waveform_report._type_hints import (
    EventType,
    IqInfoType,
    PortToDucMap,
    ChirpInfoType,
    AdcAcquisitionType,
    PlayedWaveformType,
    PulserLocationType,
    WaveformReportType,
    PlayedAnalogWaveformType,
)

logger = logging.getLogger(__name__)


class HasControllerProtocol(Protocol):
    @property
    def controller(self) -> str:
        raise NotImplementedError

    @property
    def fem(self) -> int:
        raise NotImplementedError

    @property
    def element(self) -> str:
        raise NotImplementedError


T = TypeVar("T", bound="PlayedWaveform")


@dataclass(frozen=True)
class PlayedWaveform(metaclass=ABCMeta):
    waveform_name: str
    pulse_name: str
    length: int
    timestamp: int
    iq_info: IqInfoType
    element: str
    output_ports: List[int]
    controller: str
    pulser: Dict[str, Any]
    fem: int

    @staticmethod
    def build_initialization_dict(
        dict_description: PlayedWaveformType, formatted_attribute_dict: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        attribute_dict: Dict[str, Any]
        if formatted_attribute_dict is None:
            attribute_dict = {}
        else:
            attribute_dict = deepcopy(formatted_attribute_dict)

        attribute_dict.update(
            pulse_name=dict_description["pulseName"],
            waveform_name=dict_description["waveformName"],
            timestamp=int(dict_description["timestamp"]),
            length=int(dict_description["length"]),
            iq_info=dict_description["iqInfo"],
            element=dict_description["quantumElements"],
            output_ports=[int(p) for p in dict_description["outputPorts"]],
            pulser=dict_description["pulser"],
            controller=dict_description["pulser"]["controllerName"],
            fem=int(dict_description["pulser"].get("femId", 0)) + 1,
        )
        return attribute_dict

    @classmethod
    def from_job_dict(cls: Type[T], dict_description: PlayedWaveformType) -> T:
        return cls(**cls.build_initialization_dict(dict_description))

    @property
    def ports(self) -> List[str]:
        return [str(p) for p in self.output_ports]

    @property
    def is_iq(self) -> bool:
        return self.iq_info["isPartOfIq"]

    @property
    def is_I_pulse(self) -> bool:
        return self.iq_info["isI"]

    @property
    def get_iq_association(self) -> str:
        if not self.is_iq:
            return ""
        return "I" if self.is_I_pulse else "Q"

    @property
    def ends_at(self) -> int:
        return self.timestamp + self.length

    @abstractmethod
    def to_string(self) -> str:
        return ""

    def __str__(self) -> str:
        return self.to_string()

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    def _common_attributes_to_printable_str_list(self) -> List[str]:
        waveform_type_string = "Type="
        if self.is_iq:
            waveform_type_string += f"IQ Type ({'I' if self.iq_info['isI'] else 'Q'})"
        else:
            waveform_type_string += "Single"
        return [
            f"Waveform Name={self.waveform_name}",
            f"Pulse name={self.pulse_name.removeprefix('OriginPulseName=')}",
            f"Start Time={self.timestamp} ns",
            f"Length={self.length} ns",
            f"Element={self.element}",
            f"Output Ports={self.ports}",
            waveform_type_string,
        ]


@dataclass(frozen=True)
class PlayedAnalogWaveform(PlayedWaveform):
    current_amp_elements: List[float]
    current_dc_offset_by_port: Dict[str, float]
    current_intermediate_frequency: float
    current_frame: List[float]
    current_correction_elements: List[float]
    chirp_info: Optional[ChirpInfoType]
    current_phase: float

    @classmethod
    def from_job_dict(cls, dict_description: PlayedWaveformType) -> "PlayedAnalogWaveform":
        dict_description = cast(PlayedAnalogWaveformType, dict_description)
        pulse_chirp_info = dict_description["chirpInfo"]
        is_pulse_have_chirp = len(pulse_chirp_info["units"]) > 0 or len(pulse_chirp_info["rate"]) > 0
        formatted_attribute_list = dict(
            current_amp_elements=dict_description["currentGMatrixElements"],
            current_dc_offset_by_port=dict_description["currentDCOffsetByPort"],
            current_intermediate_frequency=dict_description["currentIntermediateFrequency"],
            current_frame=dict_description["currentFrame"],
            current_correction_elements=dict_description["currentCorrectionElements"],
            chirp_info=pulse_chirp_info if is_pulse_have_chirp else None,
            current_phase=dict_description.get("currentPhase", 0),
        )
        if dict_description.get("portToDuc") and any(d["ducs"] for d in dict_description["portToDuc"]):
            formatted_attribute_list["port_to_duc"] = PlayedMwAnalogWaveform.create_port_to_duc_mapping(
                dict_description["portToDuc"]
            )
            class_to_init: Type[PlayedAnalogWaveform] = PlayedMwAnalogWaveform
        else:
            class_to_init = PlayedAnalogWaveform
        initialization_dict = class_to_init.build_initialization_dict(dict_description, formatted_attribute_list)
        return class_to_init(**initialization_dict)

    def to_custom_string(self, show_chirp: bool = True) -> str:
        _attributes = super()._common_attributes_to_printable_str_list()
        _attributes += (
            [
                f"{k}={v if self.is_iq else v[0]}"
                for k, v in [
                    ("Amplitude Values", [format_float(f) for f in self.current_amp_elements]),
                    ("Frame Values", [format_float(f) for f in self.current_frame]),
                    ("Correction Values", [format_float(f) for f in self.current_correction_elements]),
                ]
            ]
            + [
                f"Intermediate Frequency={pretty_string_freq(self.current_intermediate_frequency)}",
                f"Current DC Offset (By output ports)={ {k: format_float(v) for k, v in self.current_dc_offset_by_port.items()} }",
                f"Current Phase={format_float(self.current_phase)},",
            ]
            + ([] if (self.chirp_info is None or not show_chirp) else [f"chirp_info={self.chirp_info}"])
        )
        s = "AnalogWaveform(" + ("\n" + len("AnalogWaveform(") * " ").join(_attributes) + ")"
        return s

    def to_string(self) -> str:
        return self.to_custom_string()


def _fix_duc_idx(duc_idx: int) -> int:
    """This one is to overcome the returned 3 and 4 from the GW, when it's fixed, this one can be removed"""
    return (duc_idx - 1) % 2 + 1


@dataclass(frozen=True)
class PlayedMwAnalogWaveform(PlayedAnalogWaveform):
    port_to_duc: Dict[int, List[int]]

    @staticmethod
    def create_port_to_duc_mapping(port_to_duc: List[PortToDucMap]) -> Dict[int, List[int]]:
        return {int(p["port"]): [_fix_duc_idx(int(d)) for d in p["ducs"]] for p in port_to_duc}

    @property
    def ports(self) -> List[str]:
        to_return = []
        for port in self.output_ports:
            if port in self.port_to_duc:
                assert (
                    len(self.port_to_duc[port]) == 1
                ), f"Number of DUCs per port must equal 1, got {len(self.port_to_duc[port])}"
                to_return.append(f"{port}-{self.port_to_duc[port][0]}")
            else:
                to_return.append(str(port))
        return to_return


@dataclass(frozen=True)
class PlayedDigitalWaveform(PlayedWaveform):
    @classmethod
    def from_job_dict(cls: Type[T], dict_description: PlayedWaveformType) -> T:
        return cls(**cls.build_initialization_dict(dict_description))

    def to_string(self) -> str:
        s = (
            "DigitalWaveform("
            + ("\n" + len("DigitalWaveform(") * " ").join(self._common_attributes_to_printable_str_list())
            + ")"
        )
        return s


@dataclass(frozen=True)
class AdcAcquisition:
    start_time: int
    end_time: int
    process: str
    pulser: PulserLocationType
    quantum_element: str
    adc_ports: List[int]
    controller: str
    fem: int
    element: str

    @classmethod
    def from_job_dict(cls, dict_description: AdcAcquisitionType) -> "AdcAcquisition":
        return cls(
            start_time=int(dict_description["startTime"]),
            end_time=int(dict_description["endTime"]),
            process=dict_description["process"],
            pulser=dict_description["pulser"],
            quantum_element=dict_description["quantumElement"],
            adc_ports=[int(p) + 1 for p in dict_description["adc"]],
            controller=dict_description["pulser"]["controllerName"],
            fem=int(dict_description["pulser"].get("femId", 0)) + 1,
            element=dict_description["quantumElement"],
        )

    @property
    def ports(self) -> List[str]:
        return [str(p) for p in self.adc_ports]

    def to_string(self) -> str:
        return (
            "AdcAcquisition("
            + ("\n" + len("AdcAcquisition(") * " ").join(
                [
                    f"start_time={self.start_time}",
                    f"end_time={self.end_time}",
                    f"process={self.process}",
                    f"element={self.quantum_element}",
                    f"input_ports={self.adc_ports}",
                ]
            )
            + ")"
        )

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)


@dataclass(frozen=True)
class Event:
    name: str
    timestamp: int
    controller: str
    fem: int
    element: str
    is_i: bool
    is_q: bool

    @staticmethod
    def is_supported(dict_description: EventType) -> bool:
        supported_events_list = ["phase_reset", "reset_timestamp"]
        event_message = dict_description["eventMessage"]
        ret = event_message in supported_events_list
        if not ret:
            logger.debug(
                f"Event message {event_message} not supported in supported events list: {supported_events_list}"
            )
        return ret

    @classmethod
    def from_job_dict(cls, dict_description: EventType) -> "Event":
        return cls(
            name=dict_description["eventMessage"],
            timestamp=int(dict_description["timestamp"]),
            controller=dict_description["sourcePulser"]["controllerName"],
            fem=int(dict_description["sourcePulser"].get("femId", 0)) + 1,
            element=dict_description["quantumElement"],
            is_i=dict_description["sourcePulserIqInfo"]["isI"],
            is_q=dict_description["sourcePulserIqInfo"]["isQ"],
        )

    def to_string(self) -> str:
        """Prints all fields of the event in the standard readable format."""
        indent = " " * len("Event(")
        body = ("\n" + indent).join(f"{f.name}={getattr(self, f.name)}" for f in dataclasses.fields(self))
        return f"Event({body})"

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)


@dataclass
class FemToWaveformMap:
    analog_out: Dict[str, List[PlayedAnalogWaveform]]
    digital_out: Dict[str, List[PlayedDigitalWaveform]]
    analog_in: Dict[str, List[AdcAcquisition]]


class _SingleControllerMapping(Dict[int, FemToWaveformMap]):
    @property
    def num_analog_out_ports(self) -> int:
        return len(self.flat_analog_out)

    @property
    def num_digital_out_ports(self) -> int:
        return len(self.flat_digital_out)

    @property
    def num_analog_in_ports(self) -> int:
        return len(self.flat_analog_in)

    @property
    def flat_analog_out(self) -> Dict[str, List[PlayedAnalogWaveform]]:
        return {f"{fem_idx}-{port_idx}": v for fem_idx, fem in self.items() for port_idx, v in fem.analog_out.items()}

    @property
    def flat_digital_out(self) -> Dict[str, List[PlayedDigitalWaveform]]:
        return {f"{fem_idx}-{port_idx}": v for fem_idx, fem in self.items() for port_idx, v in fem.digital_out.items()}

    @property
    def flat_analog_in(self) -> Dict[str, List[AdcAcquisition]]:
        return {f"{fem_idx}-{port_idx}": v for fem_idx, fem in self.items() for port_idx, v in fem.analog_in.items()}


@dataclass(frozen=True)
class WaveformReport:
    job_id: Union[int, str]
    analog_waveforms: List[PlayedAnalogWaveform]
    digital_waveforms: List[PlayedDigitalWaveform]
    adc_acquisitions: List[AdcAcquisition]
    events: List[Event]

    @classmethod
    def from_dict(cls, d: WaveformReportType, job_id: Union[int, str] = -1) -> "WaveformReport":
        return cls(
            analog_waveforms=[PlayedAnalogWaveform.from_job_dict(awf) for awf in d["analogWaveforms"]],
            digital_waveforms=[PlayedDigitalWaveform.from_job_dict(dwf) for dwf in d["digitalWaveforms"]],
            adc_acquisitions=[AdcAcquisition.from_job_dict(acq) for acq in d.get("adcAcquisitions", [])],
            events=[Event.from_job_dict(event) for event in d.get("events", []) if Event.is_supported(event)],
            job_id=job_id,
        )

    @property
    def waveforms(self) -> Sequence[PlayedWaveform]:
        return cast(List[PlayedWaveform], self.analog_waveforms) + cast(List[PlayedWaveform], self.digital_waveforms)

    @property
    def controllers_in_use(self) -> Sequence[str]:
        return sorted(self.fems_in_use_by_controller)

    @property
    def num_controllers_in_use(self) -> int:
        return len(self.controllers_in_use)

    @property
    def elements_in_report(self) -> Sequence[str]:
        wf_elements = {wf.element for wf in self.waveforms}
        adc_elements = {adc.element for adc in self.adc_acquisitions}
        event_elements = {event.element for event in self.events}
        return sorted(wf_elements | adc_elements | event_elements)

    @property
    def fems_in_use_by_controller(self) -> Mapping[str, Sequence[int]]:
        fems_in_use = defaultdict(set)
        for ap in self.waveforms:
            fems_in_use[ap.controller].add(ap.fem)
        for adc in self.adc_acquisitions:
            fems_in_use[adc.controller].add(adc.fem)
        for event in self.events:
            fems_in_use[event.controller].add(event.fem)
        return {k: sorted(v) for k, v in fems_in_use.items()}

    def to_string(self) -> str:
        """
        Dumps the report into a (pretty-print) string.

        return: str
        """
        waveforms_str = [wf.to_string() for wf in self.waveforms]
        adc_string = [adc.to_string() for adc in self.adc_acquisitions]
        events_str = [event.to_string() for event in self.events]
        return "\n".join(waveforms_str + adc_string + events_str)

    def _transform_report_by_func(self, func: Callable[[HasControllerProtocol], bool]) -> "WaveformReport":
        return WaveformReport(
            analog_waveforms=list(filter(func, self.analog_waveforms)),
            digital_waveforms=list(filter(func, self.digital_waveforms)),
            adc_acquisitions=list(filter(func, self.adc_acquisitions)),
            events=list(filter(func, self.events)),
            job_id=self.job_id,
        )

    def report_by_controllers(self) -> Mapping[str, "WaveformReport"]:
        def create_filter_func(controller: str) -> Callable[[HasControllerProtocol], bool]:
            return lambda r: r.controller == controller

        by_controller_map: Dict[str, "WaveformReport"] = {}
        for con_name in self.controllers_in_use:
            con_filter = create_filter_func(con_name)
            by_controller_map[con_name] = self._transform_report_by_func(con_filter)

        return by_controller_map

    def report_by_elements(self) -> Mapping[str, "WaveformReport"]:
        def create_filter_func(_element: str) -> Callable[[HasControllerProtocol], bool]:
            return lambda r: r.element == _element

        by_element_map: Dict[str, "WaveformReport"] = {}
        for element in self.elements_in_report:
            element_filter = create_filter_func(element)
            by_element_map[element] = self._transform_report_by_func(element_filter)

        return by_element_map

    def report_by_controller_and_fems(self) -> Mapping[str, Mapping[int, "WaveformReport"]]:
        def create_filter_func(controller: str, _fem: int) -> Callable[[HasControllerProtocol], bool]:
            return lambda r: r.controller == controller and r.fem == _fem

        by_controller_fem_map: Dict[str, Dict[int, "WaveformReport"]] = {}
        for con_name in self.controllers_in_use:
            by_controller_fem_map[con_name] = {}
            for fem in self.fems_in_use_by_controller[con_name]:
                fem_filter = create_filter_func(con_name, fem)
                by_controller_fem_map[con_name][fem] = self._transform_report_by_func(fem_filter)

        return by_controller_fem_map

    def to_dict(self) -> Dict[str, Any]:
        """
        Dumps the report to a dictionary containing three keys:
            "analog_waveforms", "digital_waveforms", "acd_acquisitions".
        Each key holds the list of all the associate data.

        Returns:
            dict
        """
        return {
            "analog_waveforms": [awf.to_dict() for awf in self.analog_waveforms],
            "digital_waveforms": [dwf.to_dict() for dwf in self.digital_waveforms],
            "adc_acquisitions": [acq.to_dict() for acq in self.adc_acquisitions],
            "events": [event.to_dict() for event in self.events],
        }

    def _strict_get_report_by_output_ports(self) -> Mapping[str, _SingleControllerMapping]:
        report_by_controller_and_fem = self.report_by_controller_and_fems()
        result = {}
        for controller, fem_to_report_map in report_by_controller_and_fem.items():
            per_controller = {}
            for fem, report in fem_to_report_map.items():
                analog_out, digital_out, analog_in = defaultdict(list), defaultdict(list), defaultdict(list)
                for awf in report.analog_waveforms:
                    for p in awf.ports:
                        analog_out[p].append(awf)
                for dwf in report.digital_waveforms:
                    for p in dwf.ports:
                        digital_out[p].append(dwf)
                for adc in report.adc_acquisitions:
                    for p in adc.ports:
                        analog_in[p].append(adc)
                per_controller[fem] = FemToWaveformMap(
                    analog_out=dict(analog_out), digital_out=dict(digital_out), analog_in=dict(analog_in)
                )
            result[controller] = _SingleControllerMapping(per_controller)
        return result

    def get_report_by_output_ports(
        self, on_controller: Optional[str] = None
    ) -> Union[Mapping[str, _SingleControllerMapping], _SingleControllerMapping]:
        result = self._strict_get_report_by_output_ports()
        if on_controller is None:
            if self.num_controllers_in_use == 1:
                return result[self.controllers_in_use[0]]
            else:
                return result
        else:
            return result[on_controller]

    def create_plot(
        self,
        samples: Optional[SimulatorSamples] = None,
        controllers: Optional[Sequence[str]] = None,
        plot: bool = True,
        save_path: Optional[str] = None,
    ) -> None:
        """Creates a plot describing the pulses from each element to each port.
        See arguments description for further options.

        Args:
            samples: The raw samples as generated from the simulator. If not given, the plot will be generated without it.
            controllers: list of controllers to generate the plot. Each controller output will be saved as a different
                        file. If not given, take all the controllers who participate in the program.
            plot: Show the plot at the end of this function call.
            save_path: Save the plot to the given location. None for not saving.

        Returns:
            None
        """
        if controllers is None:
            controllers = self.controllers_in_use

        report_by_controllers = self.report_by_controllers()
        for con_name, report in report_by_controllers.items():
            if con_name not in controllers:
                continue

            if samples is None:
                from qm.waveform_report._waveform_plot_builder import _WaveformPlotBuilder

                con_builder = _WaveformPlotBuilder(report, self.job_id)
            else:
                from qm.waveform_report._waveform_plot_builder import _WaveformPlotBuilderWithSamples

                con_builder = _WaveformPlotBuilderWithSamples(report, samples[con_name], self.job_id)

            if save_path is not None:
                dirname = os.path.dirname(save_path)
                filename = f"waveform_report_{con_name}_{self.job_id}"
                con_builder.save(dirname, filename)
            if plot:
                con_builder.plot()



import logging
from io import BytesIO
from typing import Tuple, Union, BinaryIO, cast

import numpy
import numpy.typing
from numpy.lib import format as _format

from qm.type_hinting.general import NumpyArray
from qm.api.models.jobs import JobNamedResultHeader
from qm.exceptions import StreamProcessingDataLossError

logger = logging.getLogger(__name__)


def _create_results_array(count_data_written: int, header: JobNamedResultHeader, data_writer: BytesIO) -> NumpyArray:
    final_shape = _get_final_shape(count_data_written, header.shape)
    writer = BytesIO()
    _write_header(writer, final_shape, header.d_type)

    data_writer.seek(0)
    for d in data_writer:
        writer.write(d)

    writer.seek(0)
    data = cast(NumpyArray, numpy.load(writer))
    return data


def _write_header(writer: BinaryIO, shape: Tuple[int, ...], d_type: object) -> None:
    corrected_dtype = _fix_unsupported_dtype(d_type)
    _format.write_array_header_2_0(writer, {"descr": corrected_dtype, "fortran_order": False, "shape": shape})  # type: ignore[no-untyped-call]


_NP_BOOL = numpy.dtype(numpy.bool_).str
_UNSUPPORTED_DTYPE = "bool8"


def _fix_unsupported_dtype(d_type: object) -> object:
    # Numpy2 stopped support for bool8, so we need to convert it to a valid bool type
    if d_type == _UNSUPPORTED_DTYPE:
        return _NP_BOOL
    if isinstance(d_type, list):
        for idx1, elem in enumerate(d_type):
            d_type[idx1] = _fix_unsupported_dtype(elem)
    return d_type


def _get_final_shape(count: int, shape: Tuple[int, ...]) -> Tuple[int, ...]:
    if count == 1:
        final_shape = shape
    else:
        if len(shape) == 1 and shape[0] == 1:
            final_shape = (count,)
        else:
            final_shape = (count,) + shape
    return final_shape


def _standardize_slice(name: str, item: Union[int, slice], header: JobNamedResultHeader) -> slice:
    if isinstance(item, int):
        return slice(item, item + 1)
    elif isinstance(item, slice):
        step = item.step
        if step != 1 and step is not None:
            raise Exception(f"Got step={step} for item named '{name}', Fetch supports step=1 or None in slices.")
        stop = header.count_so_far if item.stop is None else item.stop
        start = 0 if item.start is None else item.start
        return slice(start, stop, step)
    raise Exception(f"fetch supports only int or slice for item named '{name}'")


def assert_no_dataloss(header: JobNamedResultHeader, job_id: str) -> None:
    if header.has_dataloss:
        raise StreamProcessingDataLossError(f"Data loss detected in data for job: {job_id}")


def log_execution_errors(header: JobNamedResultHeader, name: str, log_error: bool) -> None:
    if log_error and header.has_execution_errors:
        logger.error(
            f"Runtime errors were detected for stream named '{name}'. "
            f"Please fetch the execution report using job.execution_report() for more information."
        )



from abc import ABCMeta
from typing import Any, TypeVar, Optional, Protocol, Coroutine

import betterproto

from qm.api.base_api import BaseApi, StubType
from qm.exceptions import ErrorType, QopResponseError

SuccessType = TypeVar("SuccessType", bound=betterproto.Message)


class ResponseProtocol(Protocol[SuccessType, ErrorType]):
    success: SuccessType
    error: ErrorType


class BaseApiV2(BaseApi[StubType], metaclass=ABCMeta):
    def _run(self, coroutine: Coroutine[Any, Any, ResponseProtocol[SuccessType, ErrorType]], timeout: Optional[float] = None) -> SuccessType:  # type: ignore[override]
        response = super()._run(coroutine, timeout)  # type: ignore[type-var]
        if response.is_set("success"):  # type: ignore[attr-defined]
            return response.success
        else:
            raise QopResponseError(error=response.error)



from typing import Type, Tuple, Union, Optional, overload

import betterproto
from google.protobuf.message import Message as PbMessage

from qm._loc import _get_loc
from qm.exceptions import QmQuaException
from qm.qua._scope_management.scopes import _ArbitraryScope
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.grpc.qua import QuaProgramAnyStatement, QuaProgramArbitraryStatement, QuaProgramArbitraryContextStatement
from qm.qua._dsl.proto_any_converter import (
    ProtoMessageT,
    BetterProtoMessageT,
    any_to_proto,
    proto_to_any,
    any_to_betterproto,
    betterproto_to_any,
)


def get_arbitrary_message(name: str, message: Union[betterproto.Message, PbMessage]) -> QuaProgramAnyStatement:
    """
    Create an arbitrary message statement to support non-qua messages.
    Args:
        name (str): message name, used to identify the message type.
        message (betterproto.Message/pb message): the message needs to be encapsulated.
    Returns:
        QuaProgramAnyStatement: The arbitrary message statement.
    """
    loc = _get_loc()
    betterproto_any = betterproto_to_any(message) if isinstance(message, betterproto.Message) else proto_to_any(message)
    statement = QuaProgramArbitraryStatement(
        loc=loc,
        name=name,
        data=betterproto_any,
    )
    return QuaProgramAnyStatement(arbitrary=statement)


def arbitrary(name: str, message: Union[betterproto.Message, PbMessage]) -> None:
    """arbitrary message to support non-qua messages.

    For example, this can be used to send a protobuf message as long as the client and the server know the message type.

    Args:
        name (str): message name, used to identify the message type.
        message (betterproto.Message/pb message): the message needs to be encapsulated.
    """
    statement = get_arbitrary_message(name, message)
    scopes_manager.append_statement(statement)


def parse_arbitrary(b: bytes) -> QuaProgramArbitraryStatement:
    """
    Parse an arbitrary statement from bytes.

    Args:
        b (bytes): The bytes to parse.
    Returns:
        QuaProgramArbitraryStatement: The parsed arbitrary statement.
    """
    parsed_message = QuaProgramAnyStatement().parse(b)
    if not parsed_message.is_set("arbitrary"):
        raise QmQuaException("Failed to parse arbitrary statement, expected 'arbitrary' field not found.")
    return parsed_message.arbitrary


def arbitrary_context(name: str, data: Optional[Union[betterproto.Message, PbMessage]] = None) -> _ArbitraryScope:
    """
    arbitrary context message to support non-qua contexts.

    For example, this can be used to send a protobuf message as long as the client and the server know the message type.

    Args:
        name (str): message name, used to identify the message type.
        data (Any): the message name data.
    Returns:
        Scope: An arbitrary scope that can be used to wrap statements.
    """
    if isinstance(data, PbMessage):
        data_to_context = proto_to_any(data)
    elif isinstance(data, betterproto.Message):
        data_to_context = betterproto_to_any(data)
    else:
        data_to_context = None
    return _ArbitraryScope(loc=_get_loc(), name=name, data=data_to_context)


@overload
def extract_from_arbitrary(
    statement: Union[QuaProgramArbitraryStatement, QuaProgramArbitraryContextStatement],
    message_cls: Type[BetterProtoMessageT],
) -> Tuple[str, Optional[BetterProtoMessageT]]:
    ...


@overload
def extract_from_arbitrary(
    statement: Union[QuaProgramArbitraryStatement, QuaProgramArbitraryContextStatement],
    message_cls: Type[ProtoMessageT],
) -> Tuple[str, Optional[ProtoMessageT]]:
    ...


def extract_from_arbitrary(
    statement: Union[QuaProgramArbitraryStatement, QuaProgramArbitraryContextStatement],
    message_cls: Union[Type[BetterProtoMessageT], Type[ProtoMessageT]],
) -> Tuple[str, Optional[Union[BetterProtoMessageT, ProtoMessageT]]]:
    """
    Extracts the name and data from an arbitrary statement.
    Args:
        statement (QuaProgramArbitraryStatement or QuaProgramArbitraryContextStatement): The statement to extract from.
        message_cls (Type[BetterProtoMessageT] or Type[ProtoMessageT]): The message class to unpack into.
    Returns:
        Tuple[str, Optional[Union[BetterProtoMessageT, ProtoMessageT]]]: The name and the unpacked message.
    """

    message: Union[BetterProtoMessageT, ProtoMessageT, None]
    if statement.data is None:
        message = None
    elif issubclass(message_cls, betterproto.Message):
        message = any_to_betterproto(statement.data, message_cls)
    elif issubclass(message_cls, PbMessage):
        message = any_to_proto(statement.data, message_cls)
    else:
        raise QmQuaException(
            f"unsupported arbitrary statement, support betterproto.Message or PbMessage, got {message_cls}"
        )

    return statement.name, message



from typing import Optional

from qm.api.models.capabilities import QopCaps
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.stream_processing.stream_processing import StreamType, ResultStreamSource, declare_stream


def _standardize_timestamp_label(timestamp_stream: Optional[StreamType]) -> Optional[str]:
    timestamp_label = None
    if isinstance(timestamp_stream, str):
        scope = scopes_manager.program_scope
        scope.add_used_capability(QopCaps.command_timestamps)
        timestamp_label = _declare_save(timestamp_stream).get_var_name()
    elif isinstance(timestamp_stream, ResultStreamSource):
        scopes_manager.program_scope.add_used_capability(QopCaps.command_timestamps)
        timestamp_label = timestamp_stream.get_var_name()
    return timestamp_label


def _declare_save(tag: str, add_legacy_timestamp: bool = False) -> ResultStreamSource:
    program_scope = scopes_manager.program_scope
    result_object = program_scope.declared_streams.get(tag, None)
    if result_object is None:
        result_object = declare_stream()
        program_scope.add_stream_declaration(tag, result_object)

        if add_legacy_timestamp:
            result_object.with_timestamps().save_all(tag)
        else:
            result_object.save_all(tag)
    return result_object



from typing import Union, TypeVar, Sequence

from qm.grpc.qua import QuaProgramVarRefExpression, QuaProgramAnyScalarExpression

# The public name for the QuaProgramAnyScalarExpression type.
MessageExpressionType = QuaProgramAnyScalarExpression

MessageVarType = QuaProgramVarRefExpression

T = TypeVar("T")
OneOrMore = Union[T, Sequence[T]]



import abc
import logging
from io import BytesIO
from typing import Union, Generic, TypeVar, BinaryIO, Optional

from qm.utils.async_utils import run_async
from qm.api.v2.job_result_api import JobResultApi
from qm.api.job_result_api import JobResultServiceApi
from qm.utils.general_utils import run_until_with_timeout
from qm.type_hinting.general import NumpyArray, NumpyNumber
from qm.api.models.capabilities import QopCaps, ServerCapabilities
from qm._stream_results._multiple_streams_fetcher import MultipleStreamsFetcher
from qm.api.models.jobs import DtypeType, JobStreamingState, JobResultItemSchema, JobNamedResultHeader
from qm._stream_results._utils import (
    _standardize_slice,
    assert_no_dataloss,
    log_execution_errors,
    _create_results_array,
)

logger = logging.getLogger(__name__)


# 23 days. Reduced from 1e8 due to Windows gRPC timeout limits causing server errors (https://quantum-machines.atlassian.net/browse/OPXK-25752).
VERY_LONG_TIME = 2e6

NumpyArrayOrSingleValue = Union[NumpyNumber, NumpyArray]
ReturnedT = TypeVar("ReturnedT", NumpyArray, Optional[NumpyArrayOrSingleValue])


class BaseSingleStreamFetcher(Generic[ReturnedT], metaclass=abc.ABCMeta):
    def __init__(
        self,
        schema: JobResultItemSchema,
        service: Union[JobResultServiceApi, JobResultApi],
        capabilities: ServerCapabilities,
        multiple_streams_fetcher: Optional[MultipleStreamsFetcher],
    ) -> None:
        self._schema = schema
        self._service = service
        self._has_job_streaming_state = capabilities.supports(QopCaps.job_streaming_state)
        self._multiple_streams_fetcher = multiple_streams_fetcher
        self._validate_schema()

    @property
    def _job_id(self) -> str:
        return self._service.id

    @abc.abstractmethod
    def _validate_schema(self) -> None:
        pass

    @property
    def name(self) -> str:
        """The name of the result this handle is connected to"""
        return self._schema.name

    @property
    def job_id(self) -> str:
        """The job id this result came from"""
        return self._job_id

    @property
    def expected_count(self) -> int:
        return self._schema.expected_count

    @property
    def numpy_dtype(self) -> DtypeType:
        return self._schema.dtype

    def wait_for_values(self, count: int = 1, timeout: float = VERY_LONG_TIME) -> None:
        """Wait until we know at least `count` values were processed for this named result

        Args:
            count: The number of items to wait for
            timeout: Timeout for waiting in seconds

        """
        run_until_with_timeout(
            lambda: self.count_so_far() >= count,
            timeout=timeout,
            timeout_message=f"result {self.name} was not done in time",
        )

    def wait_for_all_values(self, timeout: float = VERY_LONG_TIME) -> bool:
        """Wait until we know all values were processed for this named result

        Args:
            timeout: Timeout for waiting in seconds

        Returns:
            True if the job finished successfully and False if the job was closed before it was done.
            If the job is still running when reaching the timeout, a TimeoutError is raised.
        """

        def on_iteration() -> bool:
            header = self.get_job_state()
            return header.done or header.closed

        def on_finish() -> bool:
            header = self.get_job_state()
            return header.done

        return run_until_with_timeout(
            on_iteration_callback=on_iteration,
            on_complete_callback=on_finish,
            timeout=timeout,
            timeout_message=f"result {self.name} was not done in time",
        )

    def is_processing(self) -> bool:
        header = self.get_job_state()
        return not (header.done or header.closed)

    def count_so_far(self) -> int:
        """also `len(handle)`

        Returns:
            The number of values this result has so far
        """
        header = self._get_named_header()
        return header.count_so_far

    def __len__(self) -> int:
        return self.count_so_far()

    def has_dataloss(self) -> bool:
        """
        Returns: true if there was data loss during job execution
        """
        state = self.get_job_state()
        return state.has_dataloss

    def get_job_state(self) -> JobStreamingState:
        if self._has_job_streaming_state:
            return self._service.get_job_state()
        #  This is just for backward compatibility
        assert isinstance(self._service, JobResultServiceApi)
        return self._service.get_state_from_header(self.name, False)

    def _get_named_header(self, check_for_errors: bool = True, flat_struct: bool = False) -> JobNamedResultHeader:
        response = self._service.get_named_header(self.name, flat_struct)
        log_execution_errors(response, self.name, check_for_errors)
        return response

    def fetch_all(self, *, check_for_errors: bool = True, flat_struct: bool = False) -> ReturnedT:
        """
        Fetches all available results from the current result stream saved in server memory.
        The result stream is populated by the `save()` or `save_all()` statements.

        Args:
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: Results will have a flat structure - dimensions will be part of the shape and not of the type

        Returns:
            All available results for the stream. This can be a single numpy scalar or a numpy array, which depends
            on the stream operators that were used. For example, using a buffer or 'with_timestamps' will return a
            numpy array.
        """
        return self.fetch(
            slice(0, None),
            check_for_errors=check_for_errors,
            flat_struct=flat_struct,
        )

    @abc.abstractmethod
    def fetch(
        self,
        item: Union[int, slice],
        *,
        check_for_errors: bool = True,
        flat_struct: bool = False,
        timeout: Optional[float] = None,
    ) -> ReturnedT:
        """
        Fetches specific results from the current result stream saved in server memory.
        The result stream is populated by the `save()` or `save_all()` statements.

        Args:
            item: The index, or a slice indicating a range, of the result in the stream.
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: Results will have a flat structure - dimensions will be part of the shape and not of the type
            timeout: Timeout for waiting in seconds.

        Returns:
            The requested result from the stream. This can be a single numpy scalar or a numpy array, which depends
            on the stream operators that were used. For example, using a buffer or 'with_timestamps' will return a
            numpy array.

        Example:
            ```python
            res.fetch() # return the item in the top position
            ```
        """
        results = self.strict_fetch(item, check_for_errors=check_for_errors, flat_struct=flat_struct, timeout=timeout)
        return self._postprocess(results, flat_struct)

    @staticmethod
    @abc.abstractmethod
    def _postprocess(fetched_data: NumpyArray, flat_struct: bool) -> ReturnedT:
        pass

    def strict_fetch(
        self,
        item: Union[int, slice],
        *,
        check_for_errors: bool = True,
        flat_struct: bool = False,
        timeout: Optional[float] = None,
    ) -> NumpyArray:
        """
        Fetches specific results from the current result stream saved in server memory.
        The result stream is populated by the `save()` or `save_all()` statements.

        Args:
            item: The index, or a slice indicating a range, of the result in the stream.
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: Results will have a flat structure - dimensions will be part of the shape and not of the type
            timeout: Timeout for waiting in seconds.

        Returns:
            The requested result from the stream. This can be a single numpy scalar or a numpy array, which depends
            on the stream operators that were used. For example, using a buffer or 'with_timestamps' will return a
            numpy array.

        Raises:
            Exception: If item is not an integer or a slice object.
            StreamProcessingDataLossError: If data loss is detected in the data for the job.

        Example:
            ```python
            res.fetch(0)         #return the item in the top position
            res.fetch(1)         #return the item in position number 2
            res.fetch(slice(1,6))# return items from position 1 to position 6 (exclusive)
                                 # same as res.fetch_all()[1:6]
            ```
        """
        if self._multiple_streams_fetcher is not None:
            flat_struct_items = {self.name} if flat_struct else frozenset()
            return self._multiple_streams_fetcher.strict_fetch(
                {self.name: item}, flat_struct_items, check_for_errors, timeout
            )[self.name]
        # This class tries to use the newer API of fetching multiple stream at once.
        # If the server does not have this capability, it resorts to the older API.
        header = self._get_named_header(check_for_errors=check_for_errors, flat_struct=flat_struct)
        assert_no_dataloss(header, self._job_id)
        slicer = _standardize_slice(self.name, item, header)
        array = self._fetch_results(header, slicer.start, slicer.stop, timeout)

        return array

    def _fetch_results(
        self, header: JobNamedResultHeader, start: int, stop: int, timeout: Optional[float]
    ) -> NumpyArray:
        data_writer = BytesIO()
        count_data_written = run_async(self._add_results_to_writer(data_writer, start, stop, timeout))
        return _create_results_array(count_data_written, header, data_writer)

    async def _add_results_to_writer(
        self, data_writer: BinaryIO, start: int, stop: int, timeout: Optional[float]
    ) -> int:
        _count_data_written = 0
        async for result in self._service.get_job_named_result(self._schema.name, start, stop - start, timeout):
            data_writer.write(result.data)
            _count_data_written += result.count_of_items

        return _count_data_written


AnySingleStreamFetcher = Union[
    BaseSingleStreamFetcher[NumpyArray],
    BaseSingleStreamFetcher[Optional[NumpyArrayOrSingleValue]],
]



from typing import Union, Optional, cast

from qm.exceptions import QmInvalidSchemaError
from qm.type_hinting.general import NumpyArray
from qm._stream_results._single_stream_fetchers._base_single_stream_fetcher import BaseSingleStreamFetcher


class SingleStreamMultipleResultFetcher(BaseSingleStreamFetcher[NumpyArray]):
    """
    A handle to a result of a pipeline terminating with ``save_all``, or to a legacy ``save`` (save using a tag,
    instead of a stream)
    """

    def _validate_schema(self) -> None:
        if self._schema.is_single:
            raise QmInvalidSchemaError("expecting a multi-result schema")

    def fetch(
        self,
        item: Union[int, slice],
        *,
        check_for_errors: bool = True,
        flat_struct: bool = False,
        timeout: Optional[float] = None,
    ) -> NumpyArray:
        """
        Fetches specific results from the current result stream saved in server memory.
        The result stream is populated by the `save()` or `save_all()` statement (or by a legacy `save` statement).

        Args:
            item: The index, or a slice indicating a range, of the result in the stream.
            check_for_errors: If true, the function would also check whether run-time errors happened during the
                program execution and would write to the logger an error message.
            flat_struct: Results will have a flat structure - dimensions will be part of the shape and not of the type
            timeout: Timeout for waiting in seconds.

        Returns:
            The requested result from this stream. This can be a single numpy scalar or a numpy array, which depends
            on the stream operators that were used. For example, using a buffer or 'with_timestamps' will return a
            numpy array.

        Example:
            ```python
            res.fetch(0)         # return the item in the top position
            res.fetch(1)         # return the item in position number 2
            res.fetch(slice(1,6))# return items from position 1 to position 6 (exclusive)
                                 # same as res.fetch_all()[1:6]
            ```
        """
        return super().fetch(
            item, check_for_errors=check_for_errors, flat_struct=flat_struct, timeout=timeout
        )  # here just to keep the upper docstring.

    @staticmethod
    def _postprocess(fetched_data: NumpyArray, flat_struct: bool) -> NumpyArray:
        # When doing with_timestamps().save_all() the resulting data structure is a bit messed up, this fixes it.
        if (
            fetched_data.dtype.fields
            and fetched_data.dtype.fields.keys() == {"value"}
            and fetched_data["value"].dtype.fields  # type: ignore[call-overload]
            and fetched_data["value"].dtype.fields.keys() == {"value", "timestamp"}  # type: ignore[call-overload]
        ):
            data_to_return = cast(NumpyArray, fetched_data["value"])  # type: ignore[call-overload]
            if data_to_return.shape[0] == 1:  # In the case of a single row, return a 1D array
                return cast(NumpyArray, data_to_return[0])
            return data_to_return
        else:
            return fetched_data



import abc
from typing import Union, Optional, Sequence

from qm.type_hinting import Number
from qm.qua._expressions import QuaVariable, QuaArrayVariable
from qm.qua._dsl.measure.analog_measure_process import MeasureProcessAbstract, _add_target_len
from qm.grpc.qua import (
    QuaProgramMeasureProcess,
    QuaProgramDigitalMeasureProcess,
    QuaProgramDigitalMeasureProcessCounting,
    QuaProgramDigitalMeasureProcessRawTimeTagging,
)


class DigitalMeasureProcess(MeasureProcessAbstract, metaclass=abc.ABCMeta):
    def __init__(self, max_time: Number):
        super().__init__()
        self._max_time = max_time

    @property
    def max_time(self) -> int:
        return int(self._max_time)

    @property
    def unwrapped(self) -> QuaProgramMeasureProcess:
        return QuaProgramMeasureProcess(digital=self._digital_unwrapped)

    @property
    @abc.abstractmethod
    def _digital_unwrapped(self) -> QuaProgramDigitalMeasureProcess:
        pass


class RawTimeTagging(DigitalMeasureProcess):
    def __init__(
        self,
        element_output: str,
        target: QuaArrayVariable[int],
        targetLen: Optional[QuaVariable[int]],
        max_time: Number,
    ):
        super().__init__(max_time)
        self.target_arr = target
        self.element_output = element_output
        self.targetLen = targetLen

    @property
    def _digital_unwrapped(self) -> QuaProgramDigitalMeasureProcess:
        unwrapped_target = self.target_arr.unwrapped

        time_tagging = QuaProgramDigitalMeasureProcessRawTimeTagging(
            max_time=self.max_time,
            element_output=self.element_output,
            target=unwrapped_target,
        )
        time_tagging = _add_target_len(time_tagging, self.targetLen)

        return QuaProgramDigitalMeasureProcess(loc=self.loc, raw_time_tagging=time_tagging)


class Counting(DigitalMeasureProcess):
    def __init__(self, element_outputs: Union[str, Sequence[str]], target: QuaVariable[int], max_time: Number):
        super().__init__(max_time)
        self.target_int = target
        self.element_outputs = element_outputs

    @property
    def _digital_unwrapped(self) -> QuaProgramDigitalMeasureProcess:
        unwrapped_target = self.target_int.unwrapped

        counting = QuaProgramDigitalMeasureProcessCounting(
            max_time=self.max_time,
            target=unwrapped_target.variable,
        )
        outputs = self.element_outputs
        if isinstance(outputs, str):
            counting.element_outputs.append(outputs)
        else:
            counting.element_outputs.extend(outputs)

        return QuaProgramDigitalMeasureProcess(loc=self.loc, counting=counting)



import abc
import warnings
from typing import TYPE_CHECKING, Any, Type, Union, Generic, Literal, TypeVar, Optional, Sequence, overload

import numpy as np

from qm._loc import _get_loc
from qm.exceptions import QmQuaException
from qm.utils import deprecation_message
from qm.type_hinting.general import NumberT
from qm.serialization.expression_serializing_visitor import ExpressionSerializingVisitor
from qm.grpc.qua import (
    QuaProgramType,
    QuaProgramDirection,
    QuaProgramAnyStatement,
    QuaProgramVarDeclaration,
    QuaProgramBinaryExpression,
    QuaProgramVarRefExpression,
    QuaProgramLiteralExpression,
    QuaProgramFunctionExpression,
    QuaProgramAnyScalarExpression,
    QuaProgramBroadcastExpression,
    QuaProgramSaveStatementSource,
    QuaProgramArrayLengthExpression,
    QuaProgramArrayVarRefExpression,
    QuaProgramLibFunctionExpression,
    QuaProgramArrayCellRefExpression,
    QuaProgramStructVarRefExpression,
    QuaProgramAssignmentStatementTarget,
    QuaProgramExternalStreamDeclaration,
    QuaProgramVarDeclarationStructMember,
    QuaProgramAdvanceInputStreamStatement,
    QuaProgramExternalStreamRefExpression,
    QuaProgramBinaryExpressionBinaryOperator,
)

_ScalarExpressionType = QuaProgramAnyScalarExpression


if TYPE_CHECKING:
    from qm.qua._qua_struct import _QuaStruct


def to_literal(value: Union[bool, int, float], dtype: QuaProgramType) -> QuaProgramLiteralExpression:
    return QuaProgramLiteralExpression(value=str(value), type=dtype, loc=_get_loc())


def literal_int(value: int) -> _ScalarExpressionType:
    return QuaProgramAnyScalarExpression(literal=to_literal(value, QuaProgramType.INT))  # type: ignore[arg-type]


def literal_bool(value: bool) -> _ScalarExpressionType:
    return QuaProgramAnyScalarExpression(literal=to_literal(value, QuaProgramType.BOOL))  # type: ignore[arg-type]


def literal_real(value: float) -> _ScalarExpressionType:
    return QuaProgramAnyScalarExpression(literal=to_literal(value, QuaProgramType.REAL))  # type: ignore[arg-type]


def io(number: Literal[1, 2]) -> _ScalarExpressionType:
    return QuaProgramAnyScalarExpression(variable=QuaProgramVarRefExpression(io_number=number, loc=_get_loc()))


def io1() -> _ScalarExpressionType:
    return io(1)


def io2() -> _ScalarExpressionType:
    return io(2)


ScalarMessageType = TypeVar(
    "ScalarMessageType",
    QuaProgramVarRefExpression,
    QuaProgramLiteralExpression,
    QuaProgramBinaryExpression,
    QuaProgramArrayCellRefExpression,
    QuaProgramArrayLengthExpression,
    QuaProgramLibFunctionExpression,
    QuaProgramFunctionExpression,
    QuaProgramBroadcastExpression,
)


S = TypeVar(
    "S",
    bound=Union[
        QuaProgramArrayVarRefExpression,
        QuaProgramAnyScalarExpression,
        QuaProgramStructVarRefExpression,
        QuaProgramExternalStreamRefExpression,
    ],
)


class QuaExpression(Generic[S], metaclass=abc.ABCMeta):
    def __init__(self, expression: S):
        self._expression = expression

    def unwrap(self) -> S:
        return self._expression

    @property
    def unwrapped(self) -> S:
        return self.unwrap()

    def __str__(self) -> str:
        return ExpressionSerializingVisitor(None).serialize(self._expression)

    def __bool__(self) -> bool:
        raise QmQuaException(
            "Attempted to use a Python logical operator on a QUA variable. If you are unsure why you got this message,"
            " please see https://docs.quantum-machines.co/latest/docs/Guides/qua_ref/#boolean-operations"
        )


class QuaNumericExpression(Generic[S, NumberT], QuaExpression[S], metaclass=abc.ABCMeta):
    def __init__(self, expression: S, t: Type[NumberT]):
        super().__init__(expression)
        self._type: Type[NumberT] = t

    @property
    def dtype(self) -> Type[NumberT]:
        return self._type

    @property
    def _qua_type(self) -> QuaProgramType:
        if issubclass(self.dtype, bool):
            return QuaProgramType.BOOL  # type: ignore[return-value]
        if issubclass(self.dtype, int):
            return QuaProgramType.INT  # type: ignore[return-value]
        if issubclass(self.dtype, float):
            return QuaProgramType.REAL  # type: ignore[return-value]
        raise NotImplementedError(f"Unsupported type - {self.dtype}")

    @property
    def _is_input_stream(self) -> bool:
        return isinstance(self, InputStreamInterface)

    def empty(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaExpression.empty()",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="This function is going to be removed, as it's not needed.",
            ),
            DeprecationWarning,
        )
        return self._expression is None

    def isFixed(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaVariable.isFixed",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use: 'QuaVariable.is_fixed()' instead",
            ),
            DeprecationWarning,
        )
        return self.is_fixed()

    def isInt(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaVariable.isInt",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use: 'QuaVariable.is_int()' instead",
            ),
            DeprecationWarning,
        )
        return self.is_int()

    def isBool(self) -> bool:
        warnings.warn(
            deprecation_message(
                method="QuaVariable.isBool",
                deprecated_in="1.1.0",
                removed_in="1.2.0",
                details="use: 'QuaVariable.is_bool()' instead",
            ),
            DeprecationWarning,
        )
        return self.is_bool()

    def is_fixed(self) -> bool:
        return issubclass(self._type, float)

    def is_int(self) -> bool:
        return self._type == int

    def is_bool(self) -> bool:
        return self._type == bool


class ScalarMessageInterface(Generic[ScalarMessageType], metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def unwrapped_scalar(self) -> ScalarMessageType:
        pass


class QuaScalarExpression(
    QuaNumericExpression[QuaProgramAnyScalarExpression, NumberT],
    ScalarMessageInterface[ScalarMessageType],
    metaclass=abc.ABCMeta,
):
    """
    A class representing a QUA scalar - could be a single value (like a variable), or the result of multiple operations
    (between several expressions) that result in a Scalar value.
    """

    def _get_binary_pb_expression(
        self, other: "ScalarOfAnyType", op: QuaProgramBinaryExpressionBinaryOperator, self_is_first: bool = True
    ) -> QuaProgramAnyScalarExpression:
        other_as_exp = to_scalar_pb_expression(other)
        self_as_exp = self.unwrapped
        if self_is_first:
            left, right = self_as_exp, other_as_exp
        else:
            left, right = other_as_exp, self_as_exp

        exp = QuaProgramAnyScalarExpression(
            binary_operation=QuaProgramBinaryExpression(loc=_get_loc(), left=left, right=right, op=op)
        )
        return exp

    def _binary(
        self, other: "Scalar[NumberT]", op: QuaProgramBinaryExpressionBinaryOperator, self_is_first: bool = True
    ) -> "QuaBinaryOperation[NumberT]":
        exp = self._get_binary_pb_expression(other, op, self_is_first)
        return QuaBinaryOperation(exp, self.dtype)

    def _boolean_binary(
        self, other: "Scalar[NumberT]", op: QuaProgramBinaryExpressionBinaryOperator, self_is_first: bool = True
    ) -> "QuaBinaryOperation[bool]":
        exp = self._get_binary_pb_expression(other, op, self_is_first)
        return QuaBinaryOperation(exp, bool)

    def _shift_operation_binary(
        self, other: "Scalar[int]", op: QuaProgramBinaryExpressionBinaryOperator, self_is_first: bool = True
    ) -> "QuaBinaryOperation[NumberT]":
        exp = self._get_binary_pb_expression(other, op, self_is_first)
        return QuaBinaryOperation(exp, self.dtype)

    def __add__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.ADD)  # type: ignore[arg-type]

    def __radd__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.ADD, self_is_first=False)  # type: ignore[arg-type]

    def __sub__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.SUB)  # type: ignore[arg-type]

    def __rsub__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.SUB, self_is_first=False)  # type: ignore[arg-type]

    def __neg__(self) -> "QuaBinaryOperation[NumberT]":
        return self.cast(0) - self

    def __gt__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, QuaProgramBinaryExpressionBinaryOperator.GT)  # type: ignore[arg-type]

    def __ge__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, QuaProgramBinaryExpressionBinaryOperator.GET)  # type: ignore[arg-type]

    def __lt__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, QuaProgramBinaryExpressionBinaryOperator.LT)  # type: ignore[arg-type]

    def __le__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":
        return self._boolean_binary(other, QuaProgramBinaryExpressionBinaryOperator.LET)  # type: ignore[arg-type]

    def __eq__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[bool]":  # type: ignore[override]
        return self._boolean_binary(other, QuaProgramBinaryExpressionBinaryOperator.EQ)  # type: ignore[arg-type]

    def __mul__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.MULT)  # type: ignore[arg-type]

    def __rmul__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.MULT, self_is_first=False)  # type: ignore[arg-type]

    def __truediv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.DIV)  # type: ignore[arg-type]

    def __rtruediv__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.DIV, self_is_first=False)  # type: ignore[arg-type]

    def __lshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(other, QuaProgramBinaryExpressionBinaryOperator.SHL)  # type: ignore[arg-type]

    def __rlshift__(self, other: int) -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(other, QuaProgramBinaryExpressionBinaryOperator.SHL, self_is_first=False)  # type: ignore[arg-type]

    def __rshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(other, QuaProgramBinaryExpressionBinaryOperator.SHR)  # type: ignore[arg-type]

    def __rrshift__(self, other: int) -> "QuaBinaryOperation[NumberT]":
        return self._shift_operation_binary(other, QuaProgramBinaryExpressionBinaryOperator.SHR, self_is_first=False)  # type: ignore[arg-type]

    def __and__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.AND)  # type: ignore[arg-type]

    def __rand__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.AND, self_is_first=False)  # type: ignore[arg-type]

    def __or__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.OR)  # type: ignore[arg-type]

    def __ror__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.OR, self_is_first=False)  # type: ignore[arg-type]

    def __xor__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.XOR)  # type: ignore[arg-type]

    def __rxor__(self, other: NumberT) -> "QuaBinaryOperation[NumberT]":
        return self._binary(other, QuaProgramBinaryExpressionBinaryOperator.XOR, self_is_first=False)  # type: ignore[arg-type]

    def __invert__(self) -> "QuaBinaryOperation[NumberT]":
        return self._binary(self.cast(True), QuaProgramBinaryExpressionBinaryOperator.XOR)  # type: ignore[arg-type]

    def cast(self, n: Union[bool, int, float]) -> NumberT:
        return self.dtype(n)

    @property
    def save_statement(self) -> QuaProgramSaveStatementSource:
        raise QmQuaException("saving is not allowed for this kind of qua expression")


class QuaArrayVariable(QuaNumericExpression[QuaProgramArrayVarRefExpression, NumberT]):
    def __init__(self, name: str, t: Type[NumberT], init_value: Sequence[Union[int, bool, float]], size: int):
        super(QuaArrayVariable, self).__init__(QuaProgramArrayVarRefExpression(name=name), t)
        self._size = size
        self._init_value = [to_literal(t(val), self._qua_type) for val in init_value]

    @property
    def declaration_statement(self) -> QuaProgramVarDeclaration:
        return QuaProgramVarDeclaration(
            name=self.unwrapped.name,
            value=self._init_value,
            type=self._qua_type,
            size=self._size,
            dim=1,
            is_input_stream=self._is_input_stream,
        )

    def __getitem__(self, item: "Scalar[int]") -> "QuaArrayCell[NumberT]":
        idx_as_pb = to_scalar_pb_expression(item)
        arr = self.unwrapped
        loc = _get_loc()
        arr.loc = loc
        item_scalar_expression = QuaProgramAnyScalarExpression(
            array_cell=QuaProgramArrayCellRefExpression(array_var=arr, index=idx_as_pb, loc=loc)
        )
        return QuaArrayCell(item_scalar_expression, self.dtype)

    def length(self) -> "QuaArrayLength[int]":
        unwrapped_element = self.unwrapped
        array_exp = QuaProgramArrayLengthExpression(array=unwrapped_element)
        result = QuaProgramAnyScalarExpression(array_length=array_exp)
        return QuaArrayLength(result, int)


class QuaStructReference(QuaExpression[QuaProgramStructVarRefExpression]):
    def __init__(self, name: str):
        super().__init__(QuaProgramStructVarRefExpression(name=name, loc=_get_loc()))
        self.name = name


NSize = TypeVar("NSize", bound=int)


class QuaStructArrayVariable(QuaArrayVariable[NumberT], Generic[NumberT, NSize]):
    def __init__(
        self,
        name: str,
        t: Type[NumberT],
        size: int,
        position: int,
        struct: QuaStructReference,
    ):
        super(QuaArrayVariable, self).__init__(
            QuaProgramArrayVarRefExpression(name=name, struct_var=struct.unwrapped), t
        )
        self._size = size
        self._init_value = []
        self._position = position
        self._struct: QuaStructReference = struct

    @property
    def declaration_statement(self) -> QuaProgramVarDeclaration:
        declaration_statement = super().declaration_statement
        struct_member = QuaProgramVarDeclarationStructMember(name=self._struct.name, position=self._position)
        declaration_statement.struct_member = struct_member
        return declaration_statement


class AssignmentTargetInterface(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def assignment_statement(self) -> QuaProgramAssignmentStatementTarget:
        pass


class NotAllowedOperation(QmQuaException):
    def __init__(self, expression_type: str) -> None:
        super(NotAllowedOperation, self).__init__(
            f"In-place operations are not supported for QUA {expression_type}. Please use `assign` instead."
        )


class NotAllowedOperationVariable(NotAllowedOperation):
    def __init__(self) -> None:
        super().__init__("variable")


class NotAllowedOperationArrayCell(NotAllowedOperation):
    def __init__(self) -> None:
        super().__init__("array cell")


class QuaVariable(AssignmentTargetInterface, QuaScalarExpression[NumberT, QuaProgramVarRefExpression]):
    """A class representing a QUA scalar variable. Note that a `QuaVariable` is also a `QuaScalarExpression`."""

    def __init__(self, name: str, t: Type[NumberT], init_value: Optional[Union[int, bool, float]]):
        super(QuaScalarExpression, self).__init__(
            QuaProgramAnyScalarExpression(variable=QuaProgramVarRefExpression(name)), t
        )
        self._init_value = [to_literal(t(init_value), self._qua_type)] if init_value is not None else []

    @property
    def declaration_statement(self) -> QuaProgramVarDeclaration:
        return QuaProgramVarDeclaration(
            name=self.unwrapped.variable.name,
            value=self._init_value,
            type=self._qua_type,
            size=1,
            dim=0,
            is_input_stream=self._is_input_stream,
        )

    @property
    def unwrapped_scalar(self) -> QuaProgramVarRefExpression:
        return self.unwrapped.variable

    @property
    def save_statement(self) -> QuaProgramSaveStatementSource:
        return QuaProgramSaveStatementSource(variable=self.unwrapped_scalar)

    @property
    def assignment_statement(self) -> QuaProgramAssignmentStatementTarget:
        return QuaProgramAssignmentStatementTarget(variable=self.unwrapped_scalar)

    def __iadd__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __isub__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __imul__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __itruediv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ifloordiv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __imod__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ipow__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ilshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __irshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __iand__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ior__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable

    def __ixor__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationVariable


class QuaLiteral(QuaScalarExpression[NumberT, QuaProgramLiteralExpression]):
    @property
    def unwrapped_scalar(self) -> QuaProgramLiteralExpression:
        return self.unwrapped.literal

    @property
    def save_statement(self) -> QuaProgramSaveStatementSource:
        return QuaProgramSaveStatementSource(literal=self.unwrapped_scalar)


class QuaArrayCell(AssignmentTargetInterface, QuaScalarExpression[NumberT, QuaProgramArrayCellRefExpression]):
    """A class representing a QUA variable inside a QUA array cell."""

    @property
    def unwrapped_scalar(self) -> QuaProgramArrayCellRefExpression:
        return self.unwrapped.array_cell

    @property
    def save_statement(self) -> QuaProgramSaveStatementSource:
        return QuaProgramSaveStatementSource(array_cell=self.unwrapped_scalar)

    @property
    def assignment_statement(self) -> QuaProgramAssignmentStatementTarget:
        return QuaProgramAssignmentStatementTarget(array_cell=self.unwrapped_scalar)

    def __iadd__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __isub__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __imul__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __itruediv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ifloordiv__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __imod__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ipow__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ilshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __irshift__(self, other: "Scalar[int]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __iand__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ior__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell

    def __ixor__(self, other: "Scalar[NumberT]") -> "QuaBinaryOperation[NumberT]":
        raise NotAllowedOperationArrayCell


class QuaBinaryOperation(QuaScalarExpression[NumberT, QuaProgramBinaryExpression]):
    @property
    def unwrapped_scalar(self) -> QuaProgramBinaryExpression:
        return self.unwrapped.binary_operation


class QuaArrayLength(QuaScalarExpression[NumberT, QuaProgramArrayLengthExpression]):
    @property
    def unwrapped_scalar(self) -> QuaProgramArrayLengthExpression:
        return self.unwrapped.array_length


class QuaLibFunctionOutput(QuaScalarExpression[NumberT, QuaProgramLibFunctionExpression]):
    """A class representing the result of a QUA lib function."""

    @property
    def unwrapped_scalar(self) -> QuaProgramLibFunctionExpression:
        return self.unwrapped.lib_function


class QuaFunctionOutput(QuaScalarExpression[NumberT, QuaProgramFunctionExpression]):
    def __init__(self, function_expression: QuaProgramFunctionExpression, t: Type[NumberT]):
        super(QuaScalarExpression, self).__init__(QuaProgramAnyScalarExpression(function=function_expression), t)

    @property
    def unwrapped_scalar(self) -> QuaProgramFunctionExpression:
        return self.unwrapped.function


class QuaBroadcast(QuaScalarExpression[NumberT, QuaProgramBroadcastExpression]):
    """A class representing the result of a QUA broadcast expression."""

    def __init__(self, t: Type[NumberT], value: _ScalarExpressionType):
        super(QuaScalarExpression, self).__init__(
            QuaProgramAnyScalarExpression(broadcast=QuaProgramBroadcastExpression(value, loc=_get_loc())), t
        )

    @property
    def unwrapped_scalar(self) -> QuaProgramBroadcastExpression:
        return self.unwrapped.broadcast


StructT = TypeVar("StructT", bound="_QuaStruct")


class QuaExternalStream(Generic[StructT], QuaExpression[QuaProgramExternalStreamRefExpression], metaclass=abc.ABCMeta):
    def __init__(self, stream_id: int, struct_t: Type[StructT]):
        super(QuaExternalStream, self).__init__(
            QuaProgramExternalStreamRefExpression(stream_id=stream_id, loc=_get_loc())
        )
        self._stream_id = stream_id
        self._struct_t = struct_t

    @property
    @abc.abstractmethod
    def _direction(self) -> QuaProgramDirection:
        pass

    @property
    def declaration_statement(self) -> QuaProgramExternalStreamDeclaration:
        return QuaProgramExternalStreamDeclaration(
            stream_id=self._stream_id,
            expected_types=self._struct_t.__underlying_declarations__,
            direction=self._direction,
        )


class QuaExternalIncomingStream(QuaExternalStream[StructT]):
    @property
    def _direction(self) -> QuaProgramDirection:
        return QuaProgramDirection.INCOMING  # type: ignore[return-value]

    def receive(self, struct: StructT) -> None:
        # Alternative API to directly call `receive_from_external_stream`.
        # Importing `receive_from_external_stream` here to avoid circular imports
        from qm.qua._dsl.external_stream import receive_from_external_stream

        receive_from_external_stream(self, struct)


class QuaExternalOutgoingStream(QuaExternalStream[StructT]):
    @property
    def _direction(self) -> QuaProgramDirection:
        return QuaProgramDirection.OUTGOING  # type: ignore[return-value]

    def send(self, struct: StructT) -> None:
        # Alternative API to directly call `send_to_external_stream`.
        # Importing `send_to_external_stream` here to avoid circular imports
        from qm.qua._dsl.external_stream import send_to_external_stream

        send_to_external_stream(self, struct)


class InputStreamInterface(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def advance(self) -> QuaProgramAnyStatement:
        pass


class QuaArrayInputStream(QuaArrayVariable[NumberT], InputStreamInterface):
    """A class representing the QUA vector that will be used as an input stream from the job to the QUA program."""

    def advance(self) -> QuaProgramAnyStatement:
        return QuaProgramAnyStatement(
            advance_input_stream=QuaProgramAdvanceInputStreamStatement(loc=_get_loc(), stream_array=self.unwrapped)
        )


class QuaVariableInputStream(QuaVariable[NumberT], InputStreamInterface):
    """A class representing the QUA variable that will be used as an input stream from the job to the QUA program."""

    def advance(self) -> QuaProgramAnyStatement:
        return QuaProgramAnyStatement(
            advance_input_stream=QuaProgramAdvanceInputStreamStatement(
                loc=_get_loc(), stream_variable=self.unwrapped.variable
            )
        )


class QuaIO(AssignmentTargetInterface):
    """A class representing the QUA IO type."""

    def __init__(self, number: Literal[1, 2]):
        self._number = number

    @property
    def assignment_statement(self) -> QuaProgramAssignmentStatementTarget:
        return QuaProgramAssignmentStatementTarget(
            variable=QuaProgramVarRefExpression(io_number=self._number, loc=_get_loc())
        )


IO1 = QuaIO(1)
IO2 = QuaIO(2)


def _fix_object_data_type(obj: Any) -> Any:
    if isinstance(obj, (np.floating, np.integer, np.bool_)):
        obj_item = obj.item()
        if isinstance(obj_item, np.longdouble):
            return float(obj_item)
        else:
            return obj_item
    else:
        return obj


def to_scalar_pb_expression(value: Union["ScalarOfAnyType", QuaIO]) -> _ScalarExpressionType:
    other = _fix_object_data_type(value)
    if isinstance(other, QuaScalarExpression):
        return other.unwrapped
    if isinstance(other, bool):  # Since bool is a subtype of int, it must be before it
        return literal_bool(other)
    if isinstance(other, int):
        return literal_int(other)
    if isinstance(other, float):
        return literal_real(other)
    if other == IO1:
        return io(1)
    if other == IO2:
        return io(2)
    raise QmQuaException(f"invalid expression: '{other}' is not a scalar expression")


@overload
def create_qua_scalar_expression(value: "QuaScalar[NumberT]") -> "QuaScalar[NumberT]":
    ...


@overload
def create_qua_scalar_expression(value: NumberT) -> QuaLiteral[NumberT]:
    ...


def create_qua_scalar_expression(value: "Scalar[NumberT]") -> "QuaScalar[NumberT]":
    if isinstance(value, QuaScalarExpression):
        return value
    if not isinstance(value, (int, float)):
        value = _fix_object_data_type(value)
    if isinstance(value, bool):  # Since bool is a subtype of int, it must be before it
        return QuaLiteral(literal_bool(value), bool)
    if isinstance(value, int):
        return QuaLiteral(literal_int(value), int)
    if isinstance(value, float):
        return QuaLiteral(literal_real(value), float)
    raise NotImplementedError


class fixed(float):
    pass


QuaScalar = Union[
    QuaVariable[NumberT],
    QuaLiteral[NumberT],
    QuaBinaryOperation[NumberT],
    QuaArrayCell[NumberT],
    QuaArrayLength[NumberT],
    QuaLibFunctionOutput[NumberT],
    QuaFunctionOutput[NumberT],
    QuaBroadcast[NumberT],
]

Scalar = Union[QuaScalar[NumberT], NumberT]
"""A generic type representing the generic `NumberT` or the QUA equivalent of it."""

Vector = Union[Sequence[NumberT], QuaArrayVariable[NumberT]]
"""A generic type representing a generic array of `NumberT`, or the QUA equivalent of it."""

ScalarOfAnyType = Union[Scalar[bool], Scalar[int], Scalar[float]]
"""A type representing a scalar value in QUA, or the equivalent python type."""

VectorOfAnyType = Union[Vector[bool], Vector[int], Vector[float]]
"""A type representing a vector value in QUA, or the equivalent python type."""



from collections.abc import Mapping
from typing import Dict, Type, Union, Literal, TypeVar, cast

import betterproto

from qm.exceptions import ConfigValidationException
from qm.api.models.capabilities import OPX_FEM_IDX, QopCaps, ServerCapabilities
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.program._dict_to_pb_converter.converters.element_converter import DEFAULT_DUC_IDX
from qm.program._dict_to_pb_converter.converters.control_device_converter.analog_output_filters_converter import (
    AnalogOutputFiltersConverter,
)
from qm.type_hinting.config_types import (
    FEM_IDX,
    Band,
    Upconverter,
    LfFemConfigType,
    MwFemConfigType,
    ControllerConfigType,
    MwUpconverterConfigType,
    AnalogInputPortConfigType,
    AnalogOutputPortConfigType,
    DigitalInputPortConfigType,
    DigitalOutputPortConfigType,
    OPX1000ControllerConfigType,
    MwFemAnalogInputPortConfigType,
    MwFemAnalogOutputPortConfigType,
    AnalogOutputPortConfigTypeOctoDac,
)
from qm.grpc.qua_config import (
    QuaConfigFemTypes,
    QuaConfigDeviceDec,
    QuaConfigVoltageLevel,
    QuaConfigControllerDec,
    QuaConfigOctoDacFemDec,
    QuaConfigMicrowaveFemDec,
    QuaConfigAnalogInputPortDec,
    QuaConfigAnalogOutputPortDec,
    QuaConfigDigitalInputPortDec,
    QuaConfigDigitalOutputPortDec,
    QuaConfigUpConverterConfigDec,
    QuaConfigOctoDacAnalogOutputPortDec,
    QuaConfigDigitalInputPortDecPolarity,
    QuaConfigMicrowaveAnalogInputPortDec,
    QuaConfigMicrowaveAnalogOutputPortDec,
    QuaConfigOctoDacAnalogOutputPortDecOutputMode,
    QuaConfigOctoDacAnalogOutputPortDecSamplingRate,
    QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode,
)

AnalogOutputType = TypeVar("AnalogOutputType", QuaConfigAnalogOutputPortDec, QuaConfigOctoDacAnalogOutputPortDec)
ControllerConfigTypeVar = TypeVar(
    "ControllerConfigTypeVar", QuaConfigOctoDacFemDec, QuaConfigControllerDec, QuaConfigMicrowaveFemDec
)


class ControlDeviceConverter(
    BaseDictToPbConverter[Union[ControllerConfigType, OPX1000ControllerConfigType], QuaConfigDeviceDec]
):
    def __init__(self, capabilities: ServerCapabilities, init_mode: bool) -> None:
        super().__init__(capabilities, init_mode)
        self._filters_converter = AnalogOutputFiltersConverter(capabilities, init_mode)

    def convert(self, input_data: Union[ControllerConfigType, OPX1000ControllerConfigType]) -> QuaConfigDeviceDec:
        return self.controlling_devices_to_pb(input_data)

    def controlling_devices_to_pb(
        self, data: Union[ControllerConfigType, OPX1000ControllerConfigType]
    ) -> QuaConfigDeviceDec:
        fems: Dict[int, QuaConfigFemTypes] = {}

        if "fems" in data:
            data = cast(OPX1000ControllerConfigType, data)
            # Here we assume that we don't declare OPX as FEM
            if set(data) & {"analog", "analog_outputs", "digital_outputs", "digital_inputs"}:
                raise Exception(
                    "'analog', 'analog_outputs', 'digital_outputs' and 'digital_inputs' are not allowed when 'fems' is present"
                )
            for k, v in data["fems"].items():
                if v.get("type") == "MW":
                    fems[int(k)] = self._mw_fem_to_pb(cast(MwFemConfigType, v))
                else:
                    fems[int(k)] = self._fem_to_pb(cast(LfFemConfigType, v))

        else:
            data = cast(ControllerConfigType, data)
            fems[OPX_FEM_IDX] = self._controller_to_pb(data)

        item = QuaConfigDeviceDec(fems=fems)
        return item

    def _controller_to_pb(self, data: ControllerConfigType) -> QuaConfigFemTypes:
        cont = QuaConfigControllerDec(type=data.get("type", "opx1"))
        cont = self._set_ports_in_config(cont, data)
        return QuaConfigFemTypes(opx=cont)

    def _fem_to_pb(self, data: LfFemConfigType) -> QuaConfigFemTypes:
        cont = QuaConfigOctoDacFemDec()
        cont = self._set_ports_in_config(cont, data)
        return QuaConfigFemTypes(octo_dac=cont)

    def _mw_fem_to_pb(self, data: MwFemConfigType) -> QuaConfigFemTypes:
        cont = QuaConfigMicrowaveFemDec()
        cont = self._set_ports_in_config(cont, data)
        return QuaConfigFemTypes(microwave=cont)

    def _set_ports_in_config(
        self,
        config: ControllerConfigTypeVar,
        data: Union[ControllerConfigType, LfFemConfigType, MwFemConfigType],
    ) -> ControllerConfigTypeVar:
        if "analog_outputs" in data:
            for analog_output_idx, analog_output_data in data["analog_outputs"].items():
                int_k = int(analog_output_idx)
                if isinstance(config, QuaConfigControllerDec):
                    analog_output_data = cast(AnalogOutputPortConfigType, analog_output_data)
                    config.analog_outputs[int_k] = self.analog_output_port_to_pb(
                        analog_output_data,
                        output_type=QuaConfigAnalogOutputPortDec,
                    )
                elif isinstance(config, QuaConfigOctoDacFemDec):
                    analog_output_data = cast(AnalogOutputPortConfigTypeOctoDac, analog_output_data)
                    config.analog_outputs[int_k] = self.opx_1000_analog_output_port_to_pb(analog_output_data)
                elif isinstance(config, QuaConfigMicrowaveFemDec):
                    analog_output_data = cast(MwFemAnalogOutputPortConfigType, analog_output_data)
                    config.analog_outputs[int_k] = self.mw_fem_analog_output_to_pb(analog_output_data)
                else:
                    raise ValueError(f"Unknown config type {type(config)}")

        if "analog_inputs" in data:
            if isinstance(config, (QuaConfigControllerDec, QuaConfigOctoDacFemDec)):
                for analog_input_idx, analog_input_data in data["analog_inputs"].items():
                    analog_input_data = cast(AnalogInputPortConfigType, analog_input_data)
                    config.analog_inputs[int(analog_input_idx)] = self.analog_input_port_to_pb(analog_input_data)
                    if isinstance(config, QuaConfigControllerDec):
                        sampling_rate = config.analog_inputs[int(analog_input_idx)].sampling_rate
                        if sampling_rate != 1e9:
                            raise ConfigValidationException(
                                f"Sampling rate of {sampling_rate} is not supported for OPX"
                            )
            elif isinstance(config, QuaConfigMicrowaveFemDec):
                for analog_input_idx, analog_input_data_mw in data["analog_inputs"].items():
                    analog_input_data_mw = cast(MwFemAnalogInputPortConfigType, analog_input_data_mw)
                    config.analog_inputs[int(analog_input_idx)] = self.mw_fem_analog_input_port_to_pb(
                        analog_input_data_mw
                    )
            else:
                raise ValueError(f"Unknown config type {type(config)}")

        if "digital_outputs" in data:
            for digital_output_idx, digital_output_data in data["digital_outputs"].items():
                config.digital_outputs[int(digital_output_idx)] = self.digital_output_port_to_pb(digital_output_data)

        if "digital_inputs" in data:
            for digital_input_idx, digital_input_data in data["digital_inputs"].items():
                config.digital_inputs[int(digital_input_idx)] = self.digital_input_port_to_pb(digital_input_data)

        return config

    def analog_input_port_to_pb(self, data: AnalogInputPortConfigType) -> QuaConfigAnalogInputPortDec:
        default_schema: AnalogInputPortConfigType = {
            "offset": 0.0,
            "shareable": False,
            "gain_db": 0,
            "sampling_rate": 1e9,
        }
        data_with_defaults = self._apply_defaults(data, default_schema=default_schema)
        analog_input = QuaConfigAnalogInputPortDec(
            offset=data_with_defaults.get("offset"),
            shareable=data_with_defaults.get("shareable"),
            gain_db=data_with_defaults.get("gain_db"),
            sampling_rate=data_with_defaults.get("sampling_rate"),
        )
        return analog_input

    def mw_fem_analog_input_port_to_pb(
        self, data: MwFemAnalogInputPortConfigType
    ) -> QuaConfigMicrowaveAnalogInputPortDec:
        if self._init_mode:
            self._validate_required_fields(data, ["band", "downconverter_frequency"], "microwave analog input port")

        default_schema: MwFemAnalogInputPortConfigType = {"sampling_rate": 1e9, "gain_db": 0, "shareable": False}
        data_with_defaults = self._apply_defaults(data, default_schema)

        analog_input = QuaConfigMicrowaveAnalogInputPortDec(
            sampling_rate=data_with_defaults.get("sampling_rate"),
            gain_db=data_with_defaults.get("gain_db"),
            shareable=data_with_defaults.get("shareable"),
            band=data_with_defaults.get("band"),
        )
        if "downconverter_frequency" in data_with_defaults:
            analog_input.downconverter.frequency = data_with_defaults["downconverter_frequency"]

        return analog_input

    def _validate_invalid_sampling_rate_and_upsampling_mode(self, data: AnalogOutputPortConfigTypeOctoDac) -> None:
        # We check that a user explicitly tried to put upsampling mode (not the default value) with a non-compatible sampling rate
        if "upsampling_mode" in data and "sampling_rate" in data and data["sampling_rate"] != 1e9:
            raise ConfigValidationException("'upsampling_mode' is only relevant for 'sampling_rate' of 1GHz.")

        # A sampling rate of 1GHZ goes hand in hand with an upsampling mode, so when updating one of these values,
        # it has to be compatible with the other.
        if not self._init_mode:
            if "sampling_rate" in data and data["sampling_rate"] == 1e9 and "upsampling_mode" not in data:
                raise ConfigValidationException(
                    "'upsampling_mode' should be provided when updating 'sampling_rate' to 1GHZ."
                )

            if "upsampling_mode" in data and "sampling_rate" not in data:
                raise ConfigValidationException(
                    "'sampling_rate' of 1GHZ should be provided when updating 'upsampling_mode'."
                )

    def analog_output_port_to_pb(
        self,
        data: AnalogOutputPortConfigType,
        output_type: Type[AnalogOutputType],
    ) -> AnalogOutputType:
        default_schema: AnalogOutputPortConfigType = {"shareable": False, "offset": 0.0, "delay": 0}
        data_with_defaults = self._apply_defaults(data, default_schema)

        analog_output = output_type(
            shareable=data_with_defaults.get("shareable"), offset=data_with_defaults.get("offset")
        )

        delay = data_with_defaults.get("delay")
        if delay is not None and delay < 0:
            raise ConfigValidationException(f"analog output delay cannot be a negative value, given value: {delay}")
        analog_output.delay = delay

        if "filter" in data_with_defaults:
            analog_output.filter = self._filters_converter.convert(data_with_defaults["filter"])

        if "crosstalk" in data_with_defaults:
            if self._capabilities.supports(QopCaps.config_v2) and isinstance(
                analog_output, QuaConfigOctoDacAnalogOutputPortDec
            ):
                crosstalk_in_pb = analog_output.crosstalk_v2.value
            else:
                crosstalk_in_pb = analog_output.crosstalk

            for k, v in data_with_defaults["crosstalk"].items():
                crosstalk_in_pb[int(k)] = v

        return analog_output

    def opx_1000_analog_output_port_to_pb(
        self,
        data: AnalogOutputPortConfigTypeOctoDac,
    ) -> QuaConfigOctoDacAnalogOutputPortDec:
        item = self.analog_output_port_to_pb(data, output_type=QuaConfigOctoDacAnalogOutputPortDec)
        self._validate_invalid_sampling_rate_and_upsampling_mode(data)

        default_schema: AnalogOutputPortConfigTypeOctoDac = {
            "sampling_rate": 1e9,
            "upsampling_mode": "mw",
            "output_mode": "direct",
        }
        data_with_defaults = self._apply_defaults(data, default_schema)

        self.update_sampling_rate_enum(item, data_with_defaults)

        if "output_mode" in data_with_defaults:
            item.output_mode = QuaConfigOctoDacAnalogOutputPortDecOutputMode[data_with_defaults.get("output_mode")]

        return item

    @staticmethod
    def update_sampling_rate_enum(
        item: QuaConfigOctoDacAnalogOutputPortDec, data_with_defaults: AnalogOutputPortConfigTypeOctoDac
    ) -> None:
        """Also update the upsampling mode, as its value is tightly correlated to the sampling rate."""
        sampling_rate = data_with_defaults.get("sampling_rate")
        if sampling_rate is not None:
            if sampling_rate == 1e9:
                item.sampling_rate = QuaConfigOctoDacAnalogOutputPortDecSamplingRate.GSPS1  # type: ignore[assignment]
                item.upsampling_mode = QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode[
                    data_with_defaults["upsampling_mode"]
                ]

            elif sampling_rate == 2e9:
                item.sampling_rate = QuaConfigOctoDacAnalogOutputPortDecSamplingRate.GSPS2  # type: ignore[assignment]
                item.upsampling_mode = QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode.unset  # type: ignore[assignment]

            else:
                raise ValueError("Sampling rate should be either 1e9 or 2e9")

    def get_upconverters(
        self, data: MwFemAnalogOutputPortConfigType, data_with_defaults: MwFemAnalogOutputPortConfigType
    ) -> Union[None, Dict[int, QuaConfigUpConverterConfigDec]]:
        upconverters = cast(Dict[int, QuaConfigUpConverterConfigDec], data_with_defaults.get("upconverters"))
        if "upconverter_frequency" in data and "upconverters" in data:
            raise ConfigValidationException("Use either 'upconverter_frequency' or 'upconverters' but not both")
        if "upconverter_frequency" in data:
            upconverters = {DEFAULT_DUC_IDX: QuaConfigUpConverterConfigDec(data["upconverter_frequency"])}
        else:
            if upconverters is not None:
                upconverters = {k: self.upconverter_config_dec_to_pb(v) for k, v in upconverters.items()}
            elif upconverters is None and self._init_mode:
                raise ConfigValidationException("You should declare at least one upconverter.")

        return cast(Union[None, Dict[int, QuaConfigUpConverterConfigDec]], upconverters)

    def mw_fem_analog_output_to_pb(
        self,
        data: MwFemAnalogOutputPortConfigType,
    ) -> QuaConfigMicrowaveAnalogOutputPortDec:
        if self._init_mode:
            self._validate_required_fields(data, ["band"], "microwave analog output port")

        default_schema: MwFemAnalogOutputPortConfigType = {
            "sampling_rate": 1e9,
            "full_scale_power_dbm": -11,
            "delay": 0,
            "shareable": False,
            "upconverters": {},
        }
        data_with_defaults = self._apply_defaults(data, default_schema)

        item = QuaConfigMicrowaveAnalogOutputPortDec(
            sampling_rate=data_with_defaults.get("sampling_rate"),
            full_scale_power_dbm=data_with_defaults.get("full_scale_power_dbm"),
            band=data_with_defaults.get("band"),
            delay=data_with_defaults.get("delay"),
            shareable=data_with_defaults.get("shareable"),
        )

        upconverters = self.get_upconverters(data, data_with_defaults)
        self._set_pb_attr_config_v2(item, upconverters, "upconverters", "upconverters_v2")

        return item

    def digital_input_port_to_pb(self, data: DigitalInputPortConfigType) -> QuaConfigDigitalInputPortDec:
        if self._init_mode:
            self._validate_required_fields(data, ["threshold", "polarity", "deadtime"], "digital input port")

        default_schema: DigitalInputPortConfigType = {"shareable": False}
        data_with_defaults = self._apply_defaults(data, default_schema=default_schema)

        digital_input = QuaConfigDigitalInputPortDec(
            shareable=data_with_defaults.get("shareable"),
            threshold=data_with_defaults.get("threshold"),
            level=QuaConfigVoltageLevel.LVTTL,  # type: ignore[arg-type]
            # The user is not supposed to edit this anymore, it should always be LVTTL. Up until now the gateway just always
            # put LVTTL here, but we are moving it here because the SDK is in charge of supplying defaults.
        )

        if "polarity" in data_with_defaults:
            if data_with_defaults["polarity"].upper() == "RISING":
                digital_input.polarity = QuaConfigDigitalInputPortDecPolarity.RISING  # type: ignore[assignment]
            elif data_with_defaults["polarity"].upper() == "FALLING":
                digital_input.polarity = QuaConfigDigitalInputPortDecPolarity.FALLING  # type: ignore[assignment]
            else:
                raise ConfigValidationException(f"Invalid polarity: {data_with_defaults['polarity']}")

        if "deadtime" in data_with_defaults:
            digital_input.deadtime = data_with_defaults["deadtime"]

        return digital_input

    def digital_output_port_to_pb(self, data: DigitalOutputPortConfigType) -> QuaConfigDigitalOutputPortDec:
        default_schema: DigitalOutputPortConfigType = {"shareable": False, "inverted": False}
        data_with_defaults = self._apply_defaults(data, default_schema)

        digital_output = QuaConfigDigitalOutputPortDec(
            shareable=data_with_defaults.get("shareable"),
            inverted=data_with_defaults.get("inverted"),
            # The only currently supported level is LVTTL, so we set it always
            level=QuaConfigVoltageLevel.LVTTL,  # type: ignore[arg-type]
        )

        return digital_output

    @staticmethod
    def upconverter_config_dec_to_pb(
        data: Union[MwUpconverterConfigType, QuaConfigUpConverterConfigDec],
    ) -> QuaConfigUpConverterConfigDec:
        if isinstance(data, QuaConfigUpConverterConfigDec):
            return data
        return QuaConfigUpConverterConfigDec(frequency=data["frequency"])

    def deconvert(self, output_data: QuaConfigDeviceDec) -> Union[ControllerConfigType, OPX1000ControllerConfigType]:

        if len(output_data.fems) == 1 and 1 in output_data.fems:
            _, opx = betterproto.which_one_of(output_data.fems[1], "fem_type_one_of")
            if isinstance(opx, QuaConfigControllerDec):
                return self._deconvert_controller(opx)

        return {
            "type": "opx1000",
            "fems": {cast(FEM_IDX, fem_idx): self._deconvert_fem(fem) for fem_idx, fem in output_data.fems.items()},
        }

    def _deconvert_controller(self, data: QuaConfigControllerDec) -> ControllerConfigType:
        return {
            "type": cast(Literal["opx", "opx1"], data.type),
            "analog_outputs": self._deconvert_controller_analog_outputs(data.analog_outputs),
            "analog_inputs": _deconvert_controller_analog_inputs(data.analog_inputs, is_opx_plus_controller=True),
            "digital_outputs": _deconvert_controller_digital_outputs(data.digital_outputs),
            "digital_inputs": _deconvert_controller_digital_inputs(data.digital_inputs),
        }

    def _deconvert_fem(self, data: QuaConfigFemTypes) -> Union[LfFemConfigType, MwFemConfigType]:
        _, fem_config = betterproto.which_one_of(data, "fem_type_one_of")
        if isinstance(fem_config, QuaConfigOctoDacFemDec):
            return self._deconvert_octo_dac(fem_config)
        elif isinstance(fem_config, QuaConfigMicrowaveFemDec):
            return self._deconvert_mw_fem(fem_config)
        else:
            raise ValueError(f"Unknown FEM type - {fem_config}")

    def _deconvert_mw_fem(self, data: QuaConfigMicrowaveFemDec) -> MwFemConfigType:
        ret: MwFemConfigType = {"type": "MW"}
        if data.analog_outputs:
            ret["analog_outputs"] = self._deconvert_mw_analog_outputs(data.analog_outputs)
        if data.analog_inputs:
            ret["analog_inputs"] = _deconvert_mw_analog_inputs(data.analog_inputs)
        if data.digital_outputs:
            ret["digital_outputs"] = _deconvert_controller_digital_outputs(data.digital_outputs)
        if data.digital_inputs:
            ret["digital_inputs"] = _deconvert_controller_digital_inputs(data.digital_inputs)
        return ret

    def _deconvert_controller_analog_outputs(
        self, outputs: dict[int, QuaConfigAnalogOutputPortDec]
    ) -> Mapping[Union[int, str], AnalogOutputPortConfigType]:
        ret: Mapping[Union[int, str], AnalogOutputPortConfigType] = {
            int(name): self._deconvert_single_analog_output(data) for name, data in outputs.items()
        }
        return ret

    def _deconvert_single_analog_output(self, data: QuaConfigAnalogOutputPortDec) -> AnalogOutputPortConfigType:
        ret = cast(
            AnalogOutputPortConfigType,
            {
                "offset": data.offset,
                "delay": data.delay,
                "shareable": data.shareable,
                "filter": self._filters_converter.deconvert(data.filter),
                "crosstalk": data.crosstalk,
            },
        )
        return ret

    def _deconvert_mw_analog_outputs(
        self,
        outputs: Dict[int, QuaConfigMicrowaveAnalogOutputPortDec],
    ) -> Mapping[Union[int, str], MwFemAnalogOutputPortConfigType]:
        return {idx: self._deconvert_single_mw_analog_output(output) for idx, output in outputs.items()}

    def _deconvert_octo_dac(self, data: QuaConfigOctoDacFemDec) -> LfFemConfigType:
        ret: LfFemConfigType = {"type": "LF"}
        if data.analog_outputs:
            ret["analog_outputs"] = self._deconvert_octo_dac_fem_analog_outputs(data.analog_outputs)
        if data.analog_inputs:
            ret["analog_inputs"] = _deconvert_controller_analog_inputs(data.analog_inputs)
        if data.digital_outputs:
            ret["digital_outputs"] = _deconvert_controller_digital_outputs(data.digital_outputs)
        if data.digital_inputs:
            ret["digital_inputs"] = _deconvert_controller_digital_inputs(data.digital_inputs)
        return ret

    def _deconvert_single_octo_dac_fem_analog_output(
        self,
        data: QuaConfigOctoDacAnalogOutputPortDec,
    ) -> AnalogOutputPortConfigTypeOctoDac:
        ret = cast(
            AnalogOutputPortConfigTypeOctoDac,
            {
                "offset": data.offset,
                "delay": data.delay,
                "shareable": data.shareable,
                "filter": self._filters_converter.deconvert(data.filter),
                "crosstalk": (
                    data.crosstalk_v2.value if self._capabilities.supports(QopCaps.config_v2) else data.crosstalk
                ),
            },
        )
        if data.sampling_rate:
            ret["sampling_rate"] = {1: 1e9, 2: 2e9}[data.sampling_rate]
        if data.upsampling_mode:
            ret["upsampling_mode"] = cast(
                Literal["mw", "pulse"], QuaConfigOctoDacAnalogOutputPortDecSamplingRateMode(data.upsampling_mode).name
            )
        if data.output_mode is not None:  # We check for "is not None" because the 0 value is valid
            ret["output_mode"] = cast(
                Literal["direct", "amplified"], QuaConfigOctoDacAnalogOutputPortDecOutputMode(data.output_mode).name
            )
        return ret

    def _deconvert_octo_dac_fem_analog_outputs(
        self,
        outputs: Dict[int, QuaConfigOctoDacAnalogOutputPortDec],
    ) -> Mapping[Union[int, str], AnalogOutputPortConfigTypeOctoDac]:
        ret: Mapping[Union[int, str], AnalogOutputPortConfigTypeOctoDac] = {
            int(name): self._deconvert_single_octo_dac_fem_analog_output(data) for name, data in outputs.items()
        }
        return ret

    def _deconvert_single_mw_analog_output(
        self, data: QuaConfigMicrowaveAnalogOutputPortDec
    ) -> MwFemAnalogOutputPortConfigType:
        upconverters = (
            data.upconverters_v2.value if self._capabilities.supports(QopCaps.config_v2) else data.upconverters
        )
        ret = cast(
            MwFemAnalogOutputPortConfigType,
            {
                "sampling_rate": data.sampling_rate,
                "full_scale_power_dbm": data.full_scale_power_dbm,
                "band": cast(Band, data.band),
                "delay": data.delay,
                "shareable": data.shareable,
                "upconverters": {cast(Upconverter, k): {"frequency": v.frequency} for k, v in upconverters.items()},
            },
        )
        return ret


def _deconvert_controller_analog_inputs(
    inputs: Mapping[int, QuaConfigAnalogInputPortDec], is_opx_plus_controller: bool = False
) -> Mapping[Union[int, str], AnalogInputPortConfigType]:
    ret: Mapping[Union[int, str], AnalogInputPortConfigType] = {
        idx: _deconvert_controller_analog_input(data, is_opx_plus_controller) for idx, data in inputs.items()
    }
    return ret


def _deconvert_controller_analog_input(
    data: QuaConfigAnalogInputPortDec, is_opx_plus_controller: bool
) -> AnalogInputPortConfigType:
    sampling_rate = data.sampling_rate
    if is_opx_plus_controller and not data.sampling_rate:
        sampling_rate = 1e9  # For OPX+ controllers, the get_config always returns 0, but we know it is 1e9 (the only allowed value for OPX+)

    ret = cast(
        AnalogInputPortConfigType,
        {
            "offset": data.offset,
            "gain_db": data.gain_db if data.gain_db is not None else 0,
            "shareable": data.shareable,
            "sampling_rate": sampling_rate,
        },
    )
    return ret


def _deconvert_controller_digital_outputs(
    outputs: Dict[int, QuaConfigDigitalOutputPortDec],
) -> Mapping[Union[int, str], DigitalOutputPortConfigType]:
    return {idx: _deconvert_controller_digital_output(data) for idx, data in outputs.items()}


def _deconvert_controller_digital_output(data: QuaConfigDigitalOutputPortDec) -> DigitalOutputPortConfigType:
    to_return = cast(
        DigitalOutputPortConfigType,
        {
            "shareable": data.shareable,
            "inverted": data.inverted,
        },
    )
    return to_return


def _deconvert_controller_digital_inputs(
    inputs: Dict[int, QuaConfigDigitalInputPortDec],
) -> Mapping[Union[int, str], DigitalInputPortConfigType]:
    return {idx: _deconvert_digital_input(data) for idx, data in inputs.items()}


def _deconvert_digital_input(data: QuaConfigDigitalInputPortDec) -> DigitalInputPortConfigType:
    to_return = cast(
        DigitalInputPortConfigType,
        {
            "deadtime": data.deadtime,
            "threshold": data.threshold,
            "shareable": data.shareable,
        },
    )
    if data.polarity is not None:
        to_return["polarity"] = cast(
            Literal["RISING", "FALLING"], QuaConfigDigitalInputPortDecPolarity(data.polarity).name
        )

    return to_return


def _deconvert_mw_analog_inputs(
    inputs: Dict[int, QuaConfigMicrowaveAnalogInputPortDec],
) -> Mapping[Union[int, str], MwFemAnalogInputPortConfigType]:
    return {idx: _deconvert_single_mw_analog_input(_input) for idx, _input in inputs.items()}


def _deconvert_single_mw_analog_input(data: QuaConfigMicrowaveAnalogInputPortDec) -> MwFemAnalogInputPortConfigType:
    return cast(
        MwFemAnalogInputPortConfigType,
        {
            "band": cast(Literal[1, 2, 3], data.band),
            "shareable": data.shareable,
            "gain_db": data.gain_db,
            "sampling_rate": data.sampling_rate,
            "downconverter_frequency": data.downconverter.frequency,
        },
    )



import abc
from collections.abc import Collection
from typing import Literal, Optional, Sequence

from betterproto.lib.google.protobuf import Value, ListValue

from qm.type_hinting import Number
from qm.qua._dsl._type_hints import OneOrMore
from qm.qua._dsl.stream_processing.stream_processing_utils import create_array


class FunctionBase(metaclass=abc.ABCMeta):
    def to_proto(self) -> Value:
        return Value(list_value=ListValue(values=[Value(string_value=self._operator_name)] + list(self._args)))

    @property
    @abc.abstractmethod
    def _operator_name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def _args(self) -> Sequence[Value]:
        pass


class Average(FunctionBase):
    def __init__(self, axis: Optional[OneOrMore[Number]] = None):
        self._axis = axis

    @property
    def _operator_name(self) -> str:
        return "average"

    @property
    def _args(self) -> Sequence[Value]:
        if self._axis is None:
            return []
        else:
            if isinstance(self._axis, Collection):
                return [create_array(self._axis)]
            else:
                return [Value(string_value=str(self._axis))]


class DotProduct(FunctionBase):
    def __init__(self, vector: Sequence[Number]):
        self._vector = vector

    @property
    def _operator_name(self) -> str:
        return "dot"

    @property
    def _args(self) -> Sequence[Value]:
        return [create_array(self._vector)]


class TupleDotProduct(FunctionBase):
    @property
    def _operator_name(self) -> str:
        return "dot"

    @property
    def _args(self) -> Sequence[Value]:
        return []


class MultiplyByScalar(FunctionBase):
    def __init__(self, scalar: Number):
        self._scalar = scalar

    @property
    def _operator_name(self) -> str:
        return "smult"

    @property
    def _args(self) -> Sequence[Value]:
        return [Value(string_value=str(self._scalar))]


class MultiplyByVector(FunctionBase):
    def __init__(self, vector: Sequence[Number]):
        self._vector = vector

    @property
    def _operator_name(self) -> str:
        return "vmult"

    @property
    def _args(self) -> Sequence[Value]:
        return [create_array(self._vector)]


class TupleMultiply(FunctionBase):
    @property
    def _operator_name(self) -> str:
        return "tmult"

    @property
    def _args(self) -> Sequence[Value]:
        return []


ConvolutionMode = Literal["", "valid", "same", "full"]


class Convolution(FunctionBase):
    def __init__(self, constant_vector: Sequence[Number], mode: ConvolutionMode = ""):
        self._constant_vector = constant_vector
        self._mode = mode

    @property
    def _operator_name(self) -> str:
        return "conv"

    @property
    def _args(self) -> Sequence[Value]:
        return [Value(string_value=self._mode), create_array(self._constant_vector)]


class TupleConvolution(FunctionBase):
    def __init__(self, mode: ConvolutionMode = ""):
        self._mode = mode

    @property
    def _operator_name(self) -> str:
        return "conv"

    @property
    def _args(self) -> Sequence[Value]:
        return [Value(string_value=self._mode)]


class FFT(FunctionBase):
    def __init__(self, output: Optional[str] = None):
        self._output = output

    @property
    def _operator_name(self) -> str:
        return "fft"

    @property
    def _args(self) -> Sequence[Value]:
        if self._output is None:
            return []
        else:
            return [Value(string_value=self._output)]


class BooleanToInt(FunctionBase):
    @property
    def _operator_name(self) -> str:
        return "booleancast"

    @property
    def _args(self) -> Sequence[Value]:
        return []


class Demod(FunctionBase):
    def __init__(
        self, frequency: Number, iw_cos: OneOrMore[Number], iw_sin: OneOrMore[Number], integrate: Optional[bool] = None
    ):
        self._frequency = frequency
        self._iw_cos = self._standardize_integration_weights(iw_cos)
        self._iw_sin = self._standardize_integration_weights(iw_sin)
        self._integrate = integrate

    @staticmethod
    def _standardize_integration_weights(iw: OneOrMore[Number]) -> Value:
        if isinstance(iw, Collection):
            return create_array(iw)
        return Value(string_value=str(iw))

    @property
    def _operator_name(self) -> str:
        return "demod"

    @property
    def _args(self) -> Sequence[Value]:
        values = [Value(string_value=str(self._frequency)), self._iw_cos, self._iw_sin]
        if isinstance(self._integrate, bool):
            values.append(Value(string_value=str(int(self._integrate))))
        return values



from typing import Dict, Union, Literal, Protocol, overload

import betterproto

from qm.exceptions import InvalidConfigError
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigFemTypes,
    QuaConfigMixerDec,
    QuaConfigPulseDec,
    QuaConfigDeviceDec,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigOscillator,
    QuaConfigQuaConfigV1,
    QuaConfigQuaConfigV2,
    QuaConfigWaveformDec,
    QuaConfigOctaveConfig,
    QuaConfigControllerDec,
    QuaConfigLogicalConfig,
    QuaConfigOctoDacFemDec,
    QuaConfigPortReference,
    QuaConfigMicrowaveFemDec,
    QuaConfigAdcPortReference,
    QuaConfigControllerConfig,
    QuaConfigDacPortReference,
    QuaConfigDigitalWaveformDec,
    QuaConfigIntegrationWeightDec,
)

FemTypes = Union[QuaConfigControllerDec, QuaConfigOctoDacFemDec, QuaConfigMicrowaveFemDec]


class ControllerConfigProtocol(Protocol):
    control_devices: Dict[str, QuaConfigDeviceDec]
    mixers: Dict[str, QuaConfigMixerDec]
    octaves: Dict[str, QuaConfigOctaveConfig]


class LogicalConfigProtocol(Protocol):
    elements: Dict[str, QuaConfigElementDec]
    oscillators: Dict[str, QuaConfigOscillator]
    pulses: Dict[str, QuaConfigPulseDec]
    waveforms: Dict[str, QuaConfigWaveformDec]
    digital_waveforms: Dict[str, QuaConfigDigitalWaveformDec]
    integration_weights: Dict[str, QuaConfigIntegrationWeightDec]


# TODO: See if the functions in this file can be moved to the config models when they are created.


@overload
def _get_correct_config_part(
    part: Literal["controller"], pb_config: QuaConfig
) -> Union[QuaConfigControllerConfig, QuaConfigQuaConfigV1]:
    pass


@overload
def _get_correct_config_part(
    part: Literal["logical"], pb_config: QuaConfig
) -> Union[QuaConfigLogicalConfig, QuaConfigQuaConfigV1]:
    pass


def _get_correct_config_part(
    part: Literal["controller", "logical"], pb_config: QuaConfig
) -> Union[QuaConfigControllerConfig, QuaConfigLogicalConfig, QuaConfigQuaConfigV1]:
    _, config_version_inst = betterproto.which_one_of(pb_config, "config_version")
    if isinstance(config_version_inst, QuaConfigQuaConfigV2):
        if part == "controller":
            return config_version_inst.controller_config
        elif part == "logical":
            return config_version_inst.logical_config
    elif isinstance(config_version_inst, QuaConfigQuaConfigV1):
        return config_version_inst
    else:
        raise ValueError("Received unknown config type: " + str(type(config_version_inst)))


def get_controller_pb_config(pb_config: QuaConfig) -> ControllerConfigProtocol:
    return _get_correct_config_part("controller", pb_config)


def get_logical_pb_config(pb_config: QuaConfig) -> LogicalConfigProtocol:
    return _get_correct_config_part("logical", pb_config)


def get_fem_config_instance(fem_ref: QuaConfigFemTypes) -> FemTypes:
    _, config = betterproto.which_one_of(fem_ref, "fem_type_one_of")
    if not isinstance(config, (QuaConfigControllerDec, QuaConfigOctoDacFemDec, QuaConfigMicrowaveFemDec)):
        raise InvalidConfigError(f"FEM type {type(config)} is not supported")
    return config


def get_fem_config(
    pb_config: QuaConfig, port: Union[QuaConfigDacPortReference, QuaConfigAdcPortReference, QuaConfigPortReference]
) -> FemTypes:
    controller_config = get_controller_pb_config(pb_config)
    if port.controller not in controller_config.control_devices:
        raise InvalidConfigError("Controller not found")
    controller = controller_config.control_devices[port.controller]
    if port.fem not in controller.fems:
        raise InvalidConfigError("FEM not found")

    fem_ref = controller.fems[port.fem]
    config = get_fem_config_instance(fem_ref)
    return config


def element_has_mix_inputs(element: QuaConfigElementDec) -> bool:
    _, inputs_inst = betterproto.which_one_of(element, "element_inputs_one_of")
    return isinstance(inputs_inst, QuaConfigMixInputs)



from typing import Union, Literal, cast

from qm.exceptions import ConfigValidationException
from qm.grpc.qua_config import QuaConfigPulseDec, QuaConfigPulseDecOperation
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.type_hinting.config_types import PulseConfigType, MixWaveformConfigType, SingleWaveformConfigType


class PulseConverter(BaseDictToPbConverter[PulseConfigType, QuaConfigPulseDec]):
    def convert(self, input_data: PulseConfigType) -> QuaConfigPulseDec:
        return self.pulse_to_pb(input_data)

    @staticmethod
    def pulse_to_pb(data: PulseConfigType) -> QuaConfigPulseDec:
        pulse = QuaConfigPulseDec()

        if "length" in data:
            pulse.length = int(data["length"])

        if data["operation"] == "control":
            pulse.operation = QuaConfigPulseDecOperation.CONTROL
        elif data["operation"] == "measurement":
            pulse.operation = QuaConfigPulseDecOperation.MEASUREMENT
        else:
            raise ConfigValidationException(f"Invalid operation {data['operation']}")

        if "digital_marker" in data:
            pulse.digital_marker = data["digital_marker"]

        if "integration_weights" in data:
            for k, v in data["integration_weights"].items():
                pulse.integration_weights[k] = v

        if "waveforms" in data:
            pulse.waveforms = {k_: str(v_) for k_, v_ in data["waveforms"].items()}
        return pulse

    def deconvert(self, output_data: QuaConfigPulseDec) -> PulseConfigType:
        temp_dict: PulseConfigType = {
            "length": output_data.length,
            "waveforms": cast(Union[SingleWaveformConfigType, MixWaveformConfigType], output_data.waveforms),
            "integration_weights": output_data.integration_weights,
            "operation": cast(
                Literal["measurement", "control"],
                QuaConfigPulseDecOperation(output_data.operation).name.lower(),  # type: ignore[union-attr]
            ),
        }
        if isinstance(output_data.digital_marker, str):
            temp_dict["digital_marker"] = output_data.digital_marker
        return temp_dict
